from typing import Any, Dict, Optional

import click
import yaspin

from grid.cli import rich_click
from grid.cli.core import Experiment
from grid.cli.utilities import is_experiment


def _format_log_lines(line: Dict[str, Any]) -> str:
    """Formats a log line for the terminal"""
    # If no timestamps are returned, fill the field
    # with dashes.
    log_type = click.style(line["type"], fg="magenta")
    if not line['timestamp']:
        # Timestamps have 32 characters.
        timestamp = click.style("-" * 32, fg="green")
    else:
        timestamp = click.style(line["timestamp"], fg="green")

    return f"[{log_type}] [{timestamp}] {line['message']}"


@rich_click.command()
@rich_click.argument("experiment", help="Name of experiment you would like to fetch logs for.", type=str, nargs=1)
@click.option(
    "--show-build-logs", type=bool, help="Shows build logs if not shown by default.", default=None, is_flag=True
)
@click.option("-l", "--tail-lines", type=int, help="Number of lines to show from the end.")
def logs(experiment: str, show_build_logs: Optional[bool] = None, tail_lines: Optional[int] = None) -> None:
    """Shows stdout logs associated with some EXPERIMENT.

    This includes both build and experiment logs.
    """

    if not is_experiment(experiment):
        message = f"""
    Experiment ID not supplied, defaulting to {experiment}-exp0.
    To view all experiments for the run:

    `grid status {experiment}`
    """
        click.echo(message)
        experiment = experiment + "-exp0"

    spinner = yaspin.yaspin(text="Fetching logs ...", color="yellow")
    spinner.start()

    try:
        # get experiment with all its metadata
        experiment = Experiment(experiment)
        experiment.refresh()

        # fetch and print stdout logs
        spinner.stop()
        for line in experiment.logs(build_logs=show_build_logs, tail_lines=tail_lines):
            click.echo(_format_log_lines(line), nl=False)

        # let the user know that certain logs are only available if experiment has run
        if experiment.is_preparing:
            styled_status = click.style(experiment.status, fg="blue")
            click.echo(f"Experiment is {styled_status}. Logs will be available when experiment starts.")
            return

        if show_build_logs is False and experiment.is_queued:
            click.echo("\n\nBuild logs skipped. Use --show-build-logs to see them.\n")
            return

    # always stop spinner otherwise it'll keep running on a background thread
    finally:
        spinner.stop()



from typing import Callable, TypeVar
from functools import wraps

TCallable = TypeVar('TCallable', bound=Callable)


def affirm(*contracts: Callable) -> Callable:
    def decorator(func: TCallable) -> TCallable:
        for contract in contracts:
            # equivalent of doing @wraps(func)
            contract = wraps(func)(contract)
            func = contract(func)
        return func

    return decorator


def is_not_deleted(func: TCallable) -> TCallable:
    """Decorator which raises an exception at access time if the run has been deleted.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._is_deleted:
            raise RuntimeError('Cannot perform operation on deleted run') from None
        return func(self, *args, **kwargs)

    return wrapper


def is_created(func: TCallable) -> TCallable:
    """Decorator which raises an exception at access time if the run has already been created.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self._is_created:
            raise RuntimeError('Invalid Operation: Allowed only after run is created') from None
        return func(self, *args, **kwargs)

    return wrapper


def is_not_created(func: TCallable) -> TCallable:
    """Decorator which raises an exception at access time if the run has not been created.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._is_created:
            raise RuntimeError('Invalid Operation: Allowed only before run is created') from None
        return func(self, *args, **kwargs)

    return wrapper


def is_not_shallow(func: TCallable) -> TCallable:
    """Decorator that un-shallow the function if it is shallow
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._is_shallow:
            self._unshallow()
        return func(self, *args, **kwargs)

    return wrapper



from typing import List

from grid.cli.core import Experiment
from grid.cli.core.base import GridObject


class Run(GridObject):
    """
    Run object in Grid. Runs are collections of Experiment objects.

    Parameters
    ----------
    name: str
        Run name (not Run ID)
    """
    def __init__(self, name: str):
        # TODO this is not the right abstraction - username shouldn't be here. Revert this hack
        # User can pass runs as username:run_name to fetch other users experiments
        username = None
        split = name.split(":")
        if len(split) > 2:
            raise ValueError(f"Error while parsing {name}. Use the format <username>:<experiment-name>")
        elif len(split) == 2:
            username = split[0]
            name = split[1]
        self._username = username
        self.identifier = name
        super().__init__()

    def refresh(self) -> None:
        """
        Updates object metadata. This makes a query to Grid to fetch the
        object's latest data.
        """
        query = """
        query GetRunDetails ($runName: ID!) {
            getRuns(runName: $runName) {
                runId
                name
                description
                entrypoint
                createdAt
                startedRunningAt
                finishedAt
                clusterId
                nExperiments
                nRunning
                nFailed
                nCompleted
                nCancelled
                nQueued
                nPending
                invocationCommand
                projectId
                config {
                    compute
                }
            }
        }
        """
        result = self.client.query(query, runName=self.identifier)
        self._data = result["getRuns"][0]
        self._update_meta()

    @property
    def experiments(self) -> List[Experiment]:
        """
        List of experiments for the Run.

        Returns
        -------
        experiments: List[Experiment]
            List of Experiment instances.
        """
        # TODO This should live in Experiment object
        query = """
        query (
                $runName: ID, $username: String
            ) {
                getExperiments (runName: $runName, username: $username) {
                    experimentId
                    name
                    commitSha
                    entrypoint
                    invocationCommands
                    createdAt
                    finishedAt
                    startedRunningAt
                    desiredState
                }
            }
        """
        result = self.client.query(query, runName=self.identifier, username=self._username)
        # Skips the need for the Experiment object to reload
        # data from the backend API.
        experiments = []
        for experiment_data in result.get("getExperiments"):
            E = Experiment(experiment_data.pop("name"))
            E.identifier = experiment_data.pop("experimentId")
            E.data = experiment_data
            experiments.append(E)

        return experiments



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class Externalv1ExperimentSpec(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'actions': 'V1RunActions',
        'auto_resume': 'bool',
        'cluster_id': 'str',
        'command': 'list[str]',
        'datastores': 'list[V1DatastoreMounts]',
        'desired_state': 'V1ExperimentState',
        'dry_run': 'bool',
        'env': 'dict(str, str)',
        'image': 'Externalv1ImageSpec',
        'instance_type': 'str',
        'resources': 'V1Resources',
        'run_controller_id': 'str',
        'run_id': 'str',
        'scratch': 'list[Gridv1ScratchSpace]',
        'source_code': 'str',
        'use_spot': 'bool',
        'user_id': 'str'
    }

    attribute_map = {
        'actions': 'actions',
        'auto_resume': 'autoResume',
        'cluster_id': 'clusterId',
        'command': 'command',
        'datastores': 'datastores',
        'desired_state': 'desiredState',
        'dry_run': 'dryRun',
        'env': 'env',
        'image': 'image',
        'instance_type': 'instanceType',
        'resources': 'resources',
        'run_controller_id': 'runControllerId',
        'run_id': 'runId',
        'scratch': 'scratch',
        'source_code': 'sourceCode',
        'use_spot': 'useSpot',
        'user_id': 'userId'
    }

    def __init__(self, actions: 'V1RunActions' = None, auto_resume: 'bool' = None, cluster_id: 'str' = None, command: 'list[str]' = None, datastores: 'list[V1DatastoreMounts]' = None, desired_state: 'V1ExperimentState' = None, dry_run: 'bool' = None, env: 'dict(str, str)' = None, image: 'Externalv1ImageSpec' = None, instance_type: 'str' = None, resources: 'V1Resources' = None, run_controller_id: 'str' = None, run_id: 'str' = None, scratch: 'list[Gridv1ScratchSpace]' = None, source_code: 'str' = None, use_spot: 'bool' = None, user_id: 'str' = None, _configuration=None):  # noqa: E501
        """Externalv1ExperimentSpec - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._actions = None
        self._auto_resume = None
        self._cluster_id = None
        self._command = None
        self._datastores = None
        self._desired_state = None
        self._dry_run = None
        self._env = None
        self._image = None
        self._instance_type = None
        self._resources = None
        self._run_controller_id = None
        self._run_id = None
        self._scratch = None
        self._source_code = None
        self._use_spot = None
        self._user_id = None
        self.discriminator = None

        if actions is not None:
            self.actions = actions
        if auto_resume is not None:
            self.auto_resume = auto_resume
        if cluster_id is not None:
            self.cluster_id = cluster_id
        if command is not None:
            self.command = command
        if datastores is not None:
            self.datastores = datastores
        if desired_state is not None:
            self.desired_state = desired_state
        if dry_run is not None:
            self.dry_run = dry_run
        if env is not None:
            self.env = env
        if image is not None:
            self.image = image
        if instance_type is not None:
            self.instance_type = instance_type
        if resources is not None:
            self.resources = resources
        if run_controller_id is not None:
            self.run_controller_id = run_controller_id
        if run_id is not None:
            self.run_id = run_id
        if scratch is not None:
            self.scratch = scratch
        if source_code is not None:
            self.source_code = source_code
        if use_spot is not None:
            self.use_spot = use_spot
        if user_id is not None:
            self.user_id = user_id

    @property
    def actions(self) -> 'V1RunActions':
        """Gets the actions of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The actions of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: V1RunActions
        """
        return self._actions

    @actions.setter
    def actions(self, actions: 'V1RunActions'):
        """Sets the actions of this Externalv1ExperimentSpec.


        :param actions: The actions of this Externalv1ExperimentSpec.  # noqa: E501
        :type: V1RunActions
        """

        self._actions = actions

    @property
    def auto_resume(self) -> 'bool':
        """Gets the auto_resume of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The auto_resume of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: bool
        """
        return self._auto_resume

    @auto_resume.setter
    def auto_resume(self, auto_resume: 'bool'):
        """Sets the auto_resume of this Externalv1ExperimentSpec.


        :param auto_resume: The auto_resume of this Externalv1ExperimentSpec.  # noqa: E501
        :type: bool
        """

        self._auto_resume = auto_resume

    @property
    def cluster_id(self) -> 'str':
        """Gets the cluster_id of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The cluster_id of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id: 'str'):
        """Sets the cluster_id of this Externalv1ExperimentSpec.


        :param cluster_id: The cluster_id of this Externalv1ExperimentSpec.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def command(self) -> 'list[str]':
        """Gets the command of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The command of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: list[str]
        """
        return self._command

    @command.setter
    def command(self, command: 'list[str]'):
        """Sets the command of this Externalv1ExperimentSpec.


        :param command: The command of this Externalv1ExperimentSpec.  # noqa: E501
        :type: list[str]
        """

        self._command = command

    @property
    def datastores(self) -> 'list[V1DatastoreMounts]':
        """Gets the datastores of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The datastores of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: list[V1DatastoreMounts]
        """
        return self._datastores

    @datastores.setter
    def datastores(self, datastores: 'list[V1DatastoreMounts]'):
        """Sets the datastores of this Externalv1ExperimentSpec.


        :param datastores: The datastores of this Externalv1ExperimentSpec.  # noqa: E501
        :type: list[V1DatastoreMounts]
        """

        self._datastores = datastores

    @property
    def desired_state(self) -> 'V1ExperimentState':
        """Gets the desired_state of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The desired_state of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: V1ExperimentState
        """
        return self._desired_state

    @desired_state.setter
    def desired_state(self, desired_state: 'V1ExperimentState'):
        """Sets the desired_state of this Externalv1ExperimentSpec.


        :param desired_state: The desired_state of this Externalv1ExperimentSpec.  # noqa: E501
        :type: V1ExperimentState
        """

        self._desired_state = desired_state

    @property
    def dry_run(self) -> 'bool':
        """Gets the dry_run of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The dry_run of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: bool
        """
        return self._dry_run

    @dry_run.setter
    def dry_run(self, dry_run: 'bool'):
        """Sets the dry_run of this Externalv1ExperimentSpec.


        :param dry_run: The dry_run of this Externalv1ExperimentSpec.  # noqa: E501
        :type: bool
        """

        self._dry_run = dry_run

    @property
    def env(self) -> 'dict(str, str)':
        """Gets the env of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The env of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._env

    @env.setter
    def env(self, env: 'dict(str, str)'):
        """Sets the env of this Externalv1ExperimentSpec.


        :param env: The env of this Externalv1ExperimentSpec.  # noqa: E501
        :type: dict(str, str)
        """

        self._env = env

    @property
    def image(self) -> 'Externalv1ImageSpec':
        """Gets the image of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The image of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: Externalv1ImageSpec
        """
        return self._image

    @image.setter
    def image(self, image: 'Externalv1ImageSpec'):
        """Sets the image of this Externalv1ExperimentSpec.


        :param image: The image of this Externalv1ExperimentSpec.  # noqa: E501
        :type: Externalv1ImageSpec
        """

        self._image = image

    @property
    def instance_type(self) -> 'str':
        """Gets the instance_type of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The instance_type of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: str
        """
        return self._instance_type

    @instance_type.setter
    def instance_type(self, instance_type: 'str'):
        """Sets the instance_type of this Externalv1ExperimentSpec.


        :param instance_type: The instance_type of this Externalv1ExperimentSpec.  # noqa: E501
        :type: str
        """

        self._instance_type = instance_type

    @property
    def resources(self) -> 'V1Resources':
        """Gets the resources of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The resources of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: V1Resources
        """
        return self._resources

    @resources.setter
    def resources(self, resources: 'V1Resources'):
        """Sets the resources of this Externalv1ExperimentSpec.


        :param resources: The resources of this Externalv1ExperimentSpec.  # noqa: E501
        :type: V1Resources
        """

        self._resources = resources

    @property
    def run_controller_id(self) -> 'str':
        """Gets the run_controller_id of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The run_controller_id of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: str
        """
        return self._run_controller_id

    @run_controller_id.setter
    def run_controller_id(self, run_controller_id: 'str'):
        """Sets the run_controller_id of this Externalv1ExperimentSpec.


        :param run_controller_id: The run_controller_id of this Externalv1ExperimentSpec.  # noqa: E501
        :type: str
        """

        self._run_controller_id = run_controller_id

    @property
    def run_id(self) -> 'str':
        """Gets the run_id of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The run_id of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: str
        """
        return self._run_id

    @run_id.setter
    def run_id(self, run_id: 'str'):
        """Sets the run_id of this Externalv1ExperimentSpec.


        :param run_id: The run_id of this Externalv1ExperimentSpec.  # noqa: E501
        :type: str
        """

        self._run_id = run_id

    @property
    def scratch(self) -> 'list[Gridv1ScratchSpace]':
        """Gets the scratch of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The scratch of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: list[Gridv1ScratchSpace]
        """
        return self._scratch

    @scratch.setter
    def scratch(self, scratch: 'list[Gridv1ScratchSpace]'):
        """Sets the scratch of this Externalv1ExperimentSpec.


        :param scratch: The scratch of this Externalv1ExperimentSpec.  # noqa: E501
        :type: list[Gridv1ScratchSpace]
        """

        self._scratch = scratch

    @property
    def source_code(self) -> 'str':
        """Gets the source_code of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The source_code of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: str
        """
        return self._source_code

    @source_code.setter
    def source_code(self, source_code: 'str'):
        """Sets the source_code of this Externalv1ExperimentSpec.


        :param source_code: The source_code of this Externalv1ExperimentSpec.  # noqa: E501
        :type: str
        """

        self._source_code = source_code

    @property
    def use_spot(self) -> 'bool':
        """Gets the use_spot of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The use_spot of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: bool
        """
        return self._use_spot

    @use_spot.setter
    def use_spot(self, use_spot: 'bool'):
        """Sets the use_spot of this Externalv1ExperimentSpec.


        :param use_spot: The use_spot of this Externalv1ExperimentSpec.  # noqa: E501
        :type: bool
        """

        self._use_spot = use_spot

    @property
    def user_id(self) -> 'str':
        """Gets the user_id of this Externalv1ExperimentSpec.  # noqa: E501


        :return: The user_id of this Externalv1ExperimentSpec.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id: 'str'):
        """Sets the user_id of this Externalv1ExperimentSpec.


        :param user_id: The user_id of this Externalv1ExperimentSpec.  # noqa: E501
        :type: str
        """

        self._user_id = user_id

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Externalv1ExperimentSpec, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'Externalv1ExperimentSpec') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, Externalv1ExperimentSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'Externalv1ExperimentSpec') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, Externalv1ExperimentSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1DatastoreUploadObject(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'expiry_time': 'str',
        'key': 'str',
        'part_count': 'str',
        'upload_id': 'str',
        'urls': 'list[V1PresignedUploadUrl]'
    }

    attribute_map = {
        'expiry_time': 'expiryTime',
        'key': 'key',
        'part_count': 'partCount',
        'upload_id': 'uploadId',
        'urls': 'urls'
    }

    def __init__(self, expiry_time: 'str' = None, key: 'str' = None, part_count: 'str' = None, upload_id: 'str' = None, urls: 'list[V1PresignedUploadUrl]' = None, _configuration=None):  # noqa: E501
        """V1DatastoreUploadObject - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._expiry_time = None
        self._key = None
        self._part_count = None
        self._upload_id = None
        self._urls = None
        self.discriminator = None

        if expiry_time is not None:
            self.expiry_time = expiry_time
        if key is not None:
            self.key = key
        if part_count is not None:
            self.part_count = part_count
        if upload_id is not None:
            self.upload_id = upload_id
        if urls is not None:
            self.urls = urls

    @property
    def expiry_time(self) -> 'str':
        """Gets the expiry_time of this V1DatastoreUploadObject.  # noqa: E501

        expiry_time refers to the Unix timestamp (rounded down to the nearest second) at which the presigned upload urls will not be available for download.  # noqa: E501

        :return: The expiry_time of this V1DatastoreUploadObject.  # noqa: E501
        :rtype: str
        """
        return self._expiry_time

    @expiry_time.setter
    def expiry_time(self, expiry_time: 'str'):
        """Sets the expiry_time of this V1DatastoreUploadObject.

        expiry_time refers to the Unix timestamp (rounded down to the nearest second) at which the presigned upload urls will not be available for download.  # noqa: E501

        :param expiry_time: The expiry_time of this V1DatastoreUploadObject.  # noqa: E501
        :type: str
        """

        self._expiry_time = expiry_time

    @property
    def key(self) -> 'str':
        """Gets the key of this V1DatastoreUploadObject.  # noqa: E501

        the relative file path/name which refers to this object.  # noqa: E501

        :return: The key of this V1DatastoreUploadObject.  # noqa: E501
        :rtype: str
        """
        return self._key

    @key.setter
    def key(self, key: 'str'):
        """Sets the key of this V1DatastoreUploadObject.

        the relative file path/name which refers to this object.  # noqa: E501

        :param key: The key of this V1DatastoreUploadObject.  # noqa: E501
        :type: str
        """

        self._key = key

    @property
    def part_count(self) -> 'str':
        """Gets the part_count of this V1DatastoreUploadObject.  # noqa: E501

        how many presigned upload URLs are needed for the object.  # noqa: E501

        :return: The part_count of this V1DatastoreUploadObject.  # noqa: E501
        :rtype: str
        """
        return self._part_count

    @part_count.setter
    def part_count(self, part_count: 'str'):
        """Sets the part_count of this V1DatastoreUploadObject.

        how many presigned upload URLs are needed for the object.  # noqa: E501

        :param part_count: The part_count of this V1DatastoreUploadObject.  # noqa: E501
        :type: str
        """

        self._part_count = part_count

    @property
    def upload_id(self) -> 'str':
        """Gets the upload_id of this V1DatastoreUploadObject.  # noqa: E501

        upload_id is the grid upload ID which is assigned to track this upload.  # noqa: E501

        :return: The upload_id of this V1DatastoreUploadObject.  # noqa: E501
        :rtype: str
        """
        return self._upload_id

    @upload_id.setter
    def upload_id(self, upload_id: 'str'):
        """Sets the upload_id of this V1DatastoreUploadObject.

        upload_id is the grid upload ID which is assigned to track this upload.  # noqa: E501

        :param upload_id: The upload_id of this V1DatastoreUploadObject.  # noqa: E501
        :type: str
        """

        self._upload_id = upload_id

    @property
    def urls(self) -> 'list[V1PresignedUploadUrl]':
        """Gets the urls of this V1DatastoreUploadObject.  # noqa: E501

        urls refers to an array of messages containing the part number and url for each presigned url part. If ``len(urls) > 0`` then this is a multi-part upload and a complete multipart-upload request call must be made to finalize the upload.  # noqa: E501

        :return: The urls of this V1DatastoreUploadObject.  # noqa: E501
        :rtype: list[V1PresignedUploadUrl]
        """
        return self._urls

    @urls.setter
    def urls(self, urls: 'list[V1PresignedUploadUrl]'):
        """Sets the urls of this V1DatastoreUploadObject.

        urls refers to an array of messages containing the part number and url for each presigned url part. If ``len(urls) > 0`` then this is a multi-part upload and a complete multipart-upload request call must be made to finalize the upload.  # noqa: E501

        :param urls: The urls of this V1DatastoreUploadObject.  # noqa: E501
        :type: list[V1PresignedUploadUrl]
        """

        self._urls = urls

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1DatastoreUploadObject, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1DatastoreUploadObject') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1DatastoreUploadObject):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1DatastoreUploadObject') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1DatastoreUploadObject):
            return True

        return self.to_dict() != other.to_dict()



import functools
import json

from grid.openapi.rest import ApiException

from typing import Callable, TypeVar


class GridException(Exception):
    """Base class for all exceptions raised by the Grid APIs"""
    pass


TCallable = TypeVar("TCallable", bound=Callable)


def throw_with_message(fn: TCallable) -> TCallable:
    """
    Decorator to ignore OpenAPI traceback and raise a custom
    exception with a message passed by Grid Backend. Strictly use it
    for all the functions in the sdk/rest module (assuming they are for
    calling Grid Backend).
    """
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ApiException as e:
            try:
                details = json.loads(e.body)['details']
            except (json.decoder.JSONDecodeError, KeyError):
                raise e
            if len(details) == 1 and isinstance(details[0], dict) and 'message' in details[0]:
                raise GridException(details[0]['message']) from None
            raise e

    return wrapper



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class Externalv1Session(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'cost': 'float',
        'created_at': 'datetime',
        'hourly_cost': 'float',
        'id': 'str',
        'name': 'str',
        'spec': 'V1SessionSpec',
        'status': 'V1SessionStatus'
    }

    attribute_map = {
        'cost': 'cost',
        'created_at': 'createdAt',
        'hourly_cost': 'hourlyCost',
        'id': 'id',
        'name': 'name',
        'spec': 'spec',
        'status': 'status'
    }

    def __init__(self, cost: 'float' = None, created_at: 'datetime' = None, hourly_cost: 'float' = None, id: 'str' = None, name: 'str' = None, spec: 'V1SessionSpec' = None, status: 'V1SessionStatus' = None, _configuration=None):  # noqa: E501
        """Externalv1Session - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._cost = None
        self._created_at = None
        self._hourly_cost = None
        self._id = None
        self._name = None
        self._spec = None
        self._status = None
        self.discriminator = None

        if cost is not None:
            self.cost = cost
        if created_at is not None:
            self.created_at = created_at
        if hourly_cost is not None:
            self.hourly_cost = hourly_cost
        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        if spec is not None:
            self.spec = spec
        if status is not None:
            self.status = status

    @property
    def cost(self) -> 'float':
        """Gets the cost of this Externalv1Session.  # noqa: E501


        :return: The cost of this Externalv1Session.  # noqa: E501
        :rtype: float
        """
        return self._cost

    @cost.setter
    def cost(self, cost: 'float'):
        """Sets the cost of this Externalv1Session.


        :param cost: The cost of this Externalv1Session.  # noqa: E501
        :type: float
        """

        self._cost = cost

    @property
    def created_at(self) -> 'datetime':
        """Gets the created_at of this Externalv1Session.  # noqa: E501


        :return: The created_at of this Externalv1Session.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at: 'datetime'):
        """Sets the created_at of this Externalv1Session.


        :param created_at: The created_at of this Externalv1Session.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def hourly_cost(self) -> 'float':
        """Gets the hourly_cost of this Externalv1Session.  # noqa: E501


        :return: The hourly_cost of this Externalv1Session.  # noqa: E501
        :rtype: float
        """
        return self._hourly_cost

    @hourly_cost.setter
    def hourly_cost(self, hourly_cost: 'float'):
        """Sets the hourly_cost of this Externalv1Session.


        :param hourly_cost: The hourly_cost of this Externalv1Session.  # noqa: E501
        :type: float
        """

        self._hourly_cost = hourly_cost

    @property
    def id(self) -> 'str':
        """Gets the id of this Externalv1Session.  # noqa: E501


        :return: The id of this Externalv1Session.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id: 'str'):
        """Sets the id of this Externalv1Session.


        :param id: The id of this Externalv1Session.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def name(self) -> 'str':
        """Gets the name of this Externalv1Session.  # noqa: E501


        :return: The name of this Externalv1Session.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: 'str'):
        """Sets the name of this Externalv1Session.


        :param name: The name of this Externalv1Session.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def spec(self) -> 'V1SessionSpec':
        """Gets the spec of this Externalv1Session.  # noqa: E501


        :return: The spec of this Externalv1Session.  # noqa: E501
        :rtype: V1SessionSpec
        """
        return self._spec

    @spec.setter
    def spec(self, spec: 'V1SessionSpec'):
        """Sets the spec of this Externalv1Session.


        :param spec: The spec of this Externalv1Session.  # noqa: E501
        :type: V1SessionSpec
        """

        self._spec = spec

    @property
    def status(self) -> 'V1SessionStatus':
        """Gets the status of this Externalv1Session.  # noqa: E501


        :return: The status of this Externalv1Session.  # noqa: E501
        :rtype: V1SessionStatus
        """
        return self._status

    @status.setter
    def status(self, status: 'V1SessionStatus'):
        """Sets the status of this Externalv1Session.


        :param status: The status of this Externalv1Session.  # noqa: E501
        :type: V1SessionStatus
        """

        self._status = status

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Externalv1Session, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'Externalv1Session') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, Externalv1Session):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'Externalv1Session') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, Externalv1Session):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2020 Grid AI Inc.
import asyncio
import csv
import json
import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional
from urllib.parse import urlsplit

import click
import requests
import websockets
import yaspin
from gql import Client, gql
from gql.transport.exceptions import TransportProtocolError, TransportQueryError
from gql.transport.requests import RequestsHTTPTransport
from gql.transport.websockets import WebsocketsTransport

from grid.cli.commands import DependencyMixin, WorkflowChecksMixin
from grid.cli.observables import Experiment, InteractiveNode, Run
from grid.cli.types import ObservableType
from grid.metadata import __version__
from grid.sdk import env

_graphql_client: Client = None

CREDIT_CARD_ERROR_MESSAGE = "A credit card on file is needed in order to use a GPU"


@dataclass
class StaticCredentials:
    user_id: str
    api_key: str


def credentials_from_env(credential_path: Optional[str] = None) -> StaticCredentials:
    """Instantiates the GraphQL local client using local credentials."""
    # if user has environment variables, use that
    env.USER_ID = os.getenv('GRID_USER_ID')
    env.API_KEY = os.getenv('GRID_API_KEY')
    if env.USER_ID and env.API_KEY:
        return StaticCredentials(user_id=env.USER_ID, api_key=env.API_KEY)

    # otherwise overwrite look for credentials stored locally as a file
    credential_path = credential_path or os.getenv(
        'GRID_CREDENTIAL_PATH',
        Path.home() / ".grid" / "credentials.json",
    )

    P = Path(credential_path)
    if not P.exists():
        raise click.ClickException('No credentials available. Did you login?')

    with P.open() as f:
        credentials = json.load(f)

    return StaticCredentials(
        user_id=credentials['UserID'],
        api_key=credentials['APIKey'],
    )


class Grid(WorkflowChecksMixin, DependencyMixin):
    """
    Interface to the Grid API.

    Attributes
    ----------
    url: str
        Grid URL
    request_timeout: int
        Number of seconds to timeout a request by default.
    client: Client
        gql client object
    grid_credentials_path: str
        Path to the Grid credentials
    default_headers: Dict[str, str]
        Header used in the request made to Grid.
    acceptable_lines_to_print: int
        Total number of acceptable lines to print in
        stdout.
    request_cooldown_duration: float
        Number of seconds to wait between continuous
        requests.

    Parameters
    ----------
    local_credentials: bool, default True
        If the client should be initialized with
        credentials from a local file or not.
    """
    url: str = env.GRID_URL
    graphql_url: str = env.GRID_URL + '/graphql'

    #  TODO: Figure out a better timeout based on query type.
    request_timeout: int = 60
    default_headers: Dict[str, str] = {'Content-type': 'application/json', 'User-Agent': f'grid-api-{__version__}'}

    grid_settings_path: str = '.grid/settings.json'
    grid_credentials_path: str = '.grid/credentials.json'

    client: Client
    transport: RequestsHTTPTransport

    available_observables: Dict[ObservableType, Callable] = {
        ObservableType.EXPERIMENT: Experiment,
        ObservableType.RUN: Run,
        ObservableType.INTERACTIVE: InteractiveNode
    }

    acceptable_lines_to_print: int = 50
    request_cooldown_duration: int = 0.1
    credentials: StaticCredentials

    def __init__(
        self,
        credential_path: Optional[str] = None,
        load_local_credentials: bool = True,
        set_context_if_not_exists: bool = True
    ):
        self.headers = self.default_headers.copy()

        #  By default, we instantiate the client with a local
        #  set of credentials.
        if load_local_credentials or credential_path:
            self._set_local_credentials(credential_path)

            #  The client will be created with a set of credentials.
            #  If we change these credentials in the context of a
            #  call, for instance "login()" then we have to
            #  re-instantiate these credentials.
            self._init_client()
        super().__init__()

    def _set_local_credentials(self, credentials_path: Optional[str] = None):
        if credentials_path:
            self.credentials = credentials_from_env(credentials_path)
        else:
            self.credentials = credentials_from_env()
        self.__set_authentication_headers(user_id=self.credentials.user_id, key=self.credentials.api_key)

    def __set_authentication_headers(self, user_id: str, key: str) -> None:
        """Sets credentials header for a client."""
        self.user_id = user_id
        self.api_key = key
        self.headers['X-Grid-User'] = user_id
        self.headers['X-Grid-Key'] = key

    def _init_client(self, websocket: bool = False) -> None:
        """
        Initializes GraphQL client. This fetches the latest
        schema from Grid.
        """
        # Check version compatibility on client initialization.
        # TODO re-enable when versioning is fixed
        # self._check_version_compatibility()

        # Print non-default
        if self.url != env.GRID_URL:
            print(f"Grid URL: {self.url}")

        if websocket:
            _url = self.url.replace('http://', 'ws://')
            _url = _url.replace('https://', 'wss://')
            _url = _url + '/subscriptions'
            self.transport = WebsocketsTransport(url=_url, init_payload=self.headers)
        else:
            self.transport = RequestsHTTPTransport(
                url=self.graphql_url, use_json=True, headers=self.headers, timeout=self.request_timeout, retries=3
            )

        try:
            # Only create a client instance if there isn't one.
            # We also check if the instantiated transport is different.
            # We'll create a new client if it is.
            global _graphql_client
            if not isinstance(_graphql_client,
                              Client) or not isinstance(_graphql_client.transport, type(self.transport)):
                _graphql_client = Client(transport=self.transport, fetch_schema_from_transport=True)

                # Initiating the context to trigger the schema fetch. This was happening
                # on the class init previously (on the above Client(...)).
                # Probably a version upgrade screwed it up.
                # Not investigating further since we are changing this soon # TODO
                if isinstance(self.transport, RequestsHTTPTransport):
                    with _graphql_client:
                        pass

            self.client = _graphql_client

        except requests.exceptions.ConnectionError:
            raise click.ClickException(f'Grid is unreachable. Is Grid online at {env.GRID_URL} ?')
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                raise click.ClickException('Not authorized. Did you login?')
            if e.response.status_code == 500:
                raise click.ClickException('Grid is having issues. Please again later.')
            raise click.ClickException('We encountered an unknown error.')

        except requests.exceptions.Timeout:
            raise click.ClickException('Could not reach Grid. Are you online?')

        except TransportProtocolError:
            raise click.ClickException('Not authorized. Did you login?')

        except Exception as e:
            raise click.ClickException(f"{type(e).__name__}: {e}")

    def query(self, query: str, **kwargs) -> Dict:
        """Executes a GraphQL query against Grid's API."""
        # instantiate http transport
        if not hasattr(self, "transport") or not isinstance(self.transport, RequestsHTTPTransport):
            self._init_client(websocket=False)

        values = kwargs or {}
        try:
            result = self.client.execute(gql(query), variable_values=values)
        except TransportQueryError as e:
            raise click.ClickException(e.args)
        except Exception as e:
            raise click.ClickException(f"{type(e).__name__}: {e}")
        if 'error' in result:
            raise click.ClickException(json.dumps(result['error']))
        return result

    def subscribe(self, query: str, **kwargs) -> Dict:
        """Streams data from a GraphQL subscription"""
        # instantiate a websocket transport
        if not hasattr(self, "transport") or not isinstance(self.transport, WebsocketsTransport):
            self._init_client(websocket=True)

        values = kwargs or {}
        try:
            stream = self.client.subscribe(gql(query), variable_values=values)
            for element in stream:
                yield element

        # If connection is suddenly closed, indicate that a known
        # error happened.
        except websockets.exceptions.ConnectionClosedError:
            raise click.ClickException("Connection closed. No new messages available.")

        except websockets.exceptions.ConnectionClosedOK:
            raise click.ClickException("Connection closed. No new messages available.")

        except asyncio.exceptions.TimeoutError:
            raise click.ClickException("Timeout waiting for messages exceeded.")

        #  Raise any other errors that the backend may raise.
        except Exception as e:  # skipcq: PYL-W0703
            raise click.ClickException(str(e))

    # skipcq: PYL-W0102
    def status(
        self,
        kind: Optional[ObservableType] = None,
        identifiers: Optional[List[str]] = None,
        follow: Optional[bool] = False,
        export: Optional[str] = None,
        is_global: Optional[bool] = False
    ) -> None:
        """
        The status of an observable object in Grid. That can be a Cluster,
        a Run, or an Experiment.

        Parameters
        ----------
        kind: Optional[ObservableType], default None
            Kind of object that we should get the status from
        identifiers: List[str], default []
            Observable identifiers
        follow: bool, default False
            If we should generate a live table with results.
        export: Optional[str], default None
            What type of file results should be exported to, if any.
        is_global: Optional[bool], default False
            Returns status of resources from everyone in the team
        """
        #  We'll instantiate a websocket client when users
        #  want to follow an observable.
        if follow:
            self._init_client(websocket=True)

        kind = kind or ObservableType.RUN

        if kind == ObservableType.EXPERIMENT:
            observable = self.available_observables[kind](client=self.client, identifier=identifiers[0])

        elif kind == ObservableType.RUN:
            if not identifiers:
                observable = self.available_observables[ObservableType.RUN](client=self.client)
            else:
                #  For now, we only check the first observable.
                #  We should also check for others in the future.
                observable = self.available_observables[kind](client=self.client, identifier=identifiers[0])

        elif kind == ObservableType.INTERACTIVE:
            # Create observable.
            observable = self.available_observables[kind](client=self.client)

        elif kind == ObservableType.CLUSTER:
            raise click.BadArgumentUsage("It isn't yet possible to observe clusters.")

        else:
            raise click.BadArgumentUsage('No observable instance created.')

        if follow:
            result = observable.follow(is_global=is_global)
        else:
            result = observable.get(is_global=is_global)

        #  Save status results to a file, if the user has specified.
        if export:
            try:

                #  No need to continue if there are not results.
                if not result:
                    click.echo('\nNo run data to write to CSV file.\n')
                    return result

                #  The user may have requested a table of
                #  Runs or Experiments, use the key that is returned
                #  by the API.
                results_key = list(result.keys())[0]

                #  Initialize variables.
                path = None
                now = datetime.now()

                #  Basic format ISO 8601
                date_string = f'{now:%Y%m%dT%H%M%S}'

                if export == 'csv':
                    path = f'grid-status-{date_string}.csv'
                    with open(path, 'w') as csv_file:

                        #  We'll exclude any List or Dict from being
                        #  exported in the CSV. We do this to avoid
                        #  generating a CSV that contains JSON data.
                        #  There aren't too many negative sides to this
                        #  because the nested data isn't as relevant.
                        sample = result[results_key][0]
                        _sample = sample.copy()
                        for k, v in _sample.items():
                            if isinstance(v, (list, dict)):
                                del sample[k]  # skipcq: PTC-W0043

                        columns = sample.keys()
                        writer = csv.DictWriter(csv_file, fieldnames=columns)
                        writer.writeheader()
                        for data in result[results_key]:
                            writer.writerow({k: v for k, v in data.items() if k in columns})

                elif export == 'json':
                    path = f'grid_status-{date_string}.json'
                    with open(path, 'w') as json_file:
                        json_file.write(json.dumps(result[results_key]))

                if path:
                    click.echo(f'\nExported status to file: {path}\n')

            #  Catch possible errors when trying to create file
            #  in file system.
            except (IOError, TypeError) as e:
                if env.DEBUG:
                    click.echo(e)

                raise click.FileError('Failed to save grid status to file\n')

        return result

    @staticmethod
    def get_experiment_id(experiment_name: str) -> str:
        """
        Get experiment ID from name.

        Parameters
        ----------
        experiment_name: str
            Experiment name

        Returns
        --------
        experiment_id: str
            Experiment ID
        """
        spinner = yaspin.yaspin(text="Getting experiment info...", color="yellow")
        try:
            # TODO refactor this when we continue the work on SDK
            from grid.cli.core import Experiment as ExperimentGridObject
            spinner.start()
            exp = ExperimentGridObject(experiment_name)
            spinner.ok("✔")
        except Exception as e:
            spinner.fail("✘")
            raise click.ClickException(
                f"Could not find experiment: {e}. "
                f"If you meant to fetch experiment owned by "
                f"another person, use the format <username>:<experiment-name>"
            )
        finally:
            spinner.stop()
        return exp.identifier

    def experiment_details(self, experiment_id: str) -> Dict[str, Any]:
        """
        Get experiment details.

        Parameters
        ----------
        experiment_id: str
            Experiment ID

        Returns
        --------
        details: Dict[str, Any]
            Experiment details
        """
        # If job is queued, notify the user that logs aren't available yet
        query = gql(
            """
        query (
            $experimentId: ID!
        ) {
            getExperimentDetails(experimentId: $experimentId) {
                experimentId
                run {
                  name
                }
                status
            }
        }
        """
        )
        params = {'experimentId': experiment_id}
        result = self.client.execute(query, variable_values=params)

        return result

    def add_ssh_public_key(self, key: str, name: str):
        return self.query(
            """
            mutation (
                $publicKey: String!
                $name: String!
            ) {
            addSSHPublicKey(name: $name, publicKey: $publicKey) {
                message
                success
                id
              }
            }
        """,
            publicKey=key,
            name=name
        )

    def list_public_ssh_keys(self, limit: int) -> List[Dict[str, str]]:
        result = self.query(
            """
            query (
                $limit: Int!,
            ) {
                getPublicSSHKeys(limit: $limit) {
                    id
                    publicKey,
                    name
              }
            }
        """,
            limit=limit,
        )
        return result['getPublicSSHKeys']

    def delete_ssh_public_key(self, key_id: str):
        self.query(
            """
            mutation (
                $id: ID!
            ) {
              deleteSSHPublicKey(id: $id) {
                message
                success
              }
            }
        """,
            id=key_id
        )

    def list_interactive_node_ssh_setting(self):
        return self.query(
            """
        query {
            getInteractiveNodes {
                name: interactiveNodeName
                ssh_url: sshUrl
                status
            }
        }
        """
        )['getInteractiveNodes']

    @staticmethod
    def _gen_ssh_config(
        interactive_nodes: List[Dict[str, Any]],
        curr_config: str,
        start_marker: str = "### grid.ai managed BEGIN do not edit manually###",
        end_marker: str = "### grid.ai managed END do not edit manually###",
    ):

        content = curr_config.splitlines()
        sol = []
        managed_part = [start_marker]
        for node in interactive_nodes:
            if node["status"] != "running":
                continue

            ssh_url = urlsplit(node['ssh_url'])
            managed_part.extend([
                f"Host {node['name']}",
                f"    User {ssh_url.username}",
                f"    Hostname {ssh_url.hostname}",
                f"    Port {ssh_url.port}",
                "    StrictHostKeyChecking accept-new",
                "    CheckHostIP no",
                "    ServerAliveInterval 15",
                "    ServerAliveCountMax 4",
                # Workarounds until https://linear.app/gridai/issue/GI-6940/switch-client-ssh-gateway-communication-to-use-ed25519-keys-over-rsa
                # is fixed
                "    HostKeyAlgorithms=+ssh-rsa",
                "    PubkeyAcceptedKeyTypes +ssh-rsa",
                "    PasswordAuthentication no",
            ])
        managed_part.append(end_marker)

        within_section = False
        added_managed_part = False
        for line in content:
            if line == start_marker:
                if added_managed_part:
                    raise ValueError("Found 2 start markers")
                if within_section:
                    raise ValueError("Found 2 start markers in row")
                within_section = True
            elif end_marker == line:
                if added_managed_part:
                    raise ValueError("Found 2+ start end")
                if not within_section:
                    raise ValueError("End marker before start marker")
                within_section = False
                sol.extend(managed_part)
                added_managed_part = True
            elif not within_section:
                sol.append(line)
        if within_section:
            raise ValueError("Found only start marker, no end one found")
        if not added_managed_part:
            sol.extend(managed_part)
        return '\n'.join(sol)

    def sync_ssh_config(self) -> List[str]:
        """
        sync local ssh config with grid's interactive nodes config

        Returns
        -------
        list of interactive nodes present
        """
        interactive_nodes = self.list_interactive_node_ssh_setting()

        ssh_config = Path(env.GRID_SSH_CONFIG)
        if not ssh_config.exists():
            ssh_config.write_text("")

        ssh_config.write_text(
            self._gen_ssh_config(
                interactive_nodes=interactive_nodes,
                curr_config=ssh_config.read_text(),
            )
        )

        return interactive_nodes



from copy import copy
from pathlib import Path
from typing import Set
import click


def generate(src: Path) -> Set[str]:
    """
    Generates list of glob strings to ignore. These need to be parsed to identify
    which files are actually ignored.

    Parameters
    ----------
    src: Path
        Path to location where to generate ignore list from.

    Returns
    -------
    Set[str]
        Glob strings to parse
    """
    # fallback files in case we can't find .gridignore
    fallback_ignore_files = {".dockerignore", ".gitignore"}

    # never include these paths in the package
    excluded_paths = {".git"}

    # if it is a file, then just read from it and return lines
    if src.is_file():
        return _read_and_filter_gridignore(src)

    # ignores all paths from excluded paths by default
    ignore_globs = {f"{p}/" for p in excluded_paths}

    # look first for `.gridignore` files
    # Nested handling does not work anyway atm. We will concentrate to make local .gridignore file work first.
    # Previous solution that added relative paths to ignore patterns would not work as intended.
    # E.g. if in dir nested/ there's .gridignore with rule *.pyc it would ignore just files in nested/*.pyc and not in all sub-directories
    # Because .gridignore handling is somewhat complicated and non-trivial I suggest to just focus on level-0 .gridignore (non-nested)
    if (src / ".gridignore").exists():
        ignore_globs.update(_read_and_filter_gridignore(src / ".gridignore"))

    nested_gridignores = set(
        gridignore_file for gridignore_file in src.rglob(".gridignore") if gridignore_file != src / ".gridignore"
    )
    if nested_gridignores:
        click.echo(
            "We've found other .gridignore files in nested directories. Please note that currently nested .gridignore files are not supported and ignore file has to be in root directory."
        )

    # if found .gridignore, return it
    if len(ignore_globs) > len(excluded_paths):
        return ignore_globs

    # if not found, look everything else -- combine all fallback_ignore_files into one
    # This is a mess atm. We do not handle all cases of .gitinogre correctly anyway, should we actually do this?
    for path in src.glob('*'):
        if path.name in excluded_paths:
            continue
        if path.is_file():
            if path.name in fallback_ignore_files:
                filtered = _read_and_filter_gridignore(path)
                relative_dir = path.relative_to(src).parents[0]
                relative_globs = [str(relative_dir / glob) for glob in filtered]
                ignore_globs.update(relative_globs)

    return ignore_globs


def _read_and_filter_gridignore(path: Path) -> Set[str]:
    """
    Reads ignore file and filters empty lines.
    Atm we support a subset of behaviour described in https://git-scm.com/docs/gitignore

    Parameters
    ----------
    path: Path
        Path to .gridignore file or equivalent.

    Returns
    -------
    Set[str]
        Set of unique ignore pattern lines.
    """

    with path.open() as f:
        # Remove excess whitespaces and comments
        strip_lines = [ln.strip() for ln in f.readlines()]
        return {ln for ln in strip_lines if ln != "" and ln is not None and not ln.startswith("#")}



# TODO - this file needs to be disected and moved into the correct place
#  a lot of functions are useless once SDK is completly migrated to to use REST

from collections import OrderedDict
from pathlib import Path
from typing import Union

import yaml

DISK_SIZE_ERROR_MSG = "Invalid disk size, should be greater than 100Gb"

SPECIAL_NAME_TO_SKIP_OBJECT_INIT = (
    '486d6d2c206120637572696f7573206465762049207365652e20436865'
    '636b6f75742068747470733a2f2f7777772e6c696e6b6564696e2e636f6'
    'd2f636f6d70616e792f677269642d61692f6a6f62732f20696620796f75'
    '206172652070617373696f6e6174652061626f7574206d616b696e67204d'
    '4c4f707320626f72696e67'
)


def check_run_name_is_valid(value: str):
    """Click callback that checks if a Run contains reserved names."""
    if value is not None:
        fail = False

        #  Check if the input is alphanumeric.
        _run_name = value.replace('-', '')
        if not _run_name.isalnum():
            fail = True

        #  Check if the allowed `-` character is not used
        #  at the end of the string.
        elif value.endswith('-') or value.startswith('-'):
            fail = True

        #  Check that the run name does not contain any
        #  uppercase characters.
        elif any(x.isupper() for x in value):
            fail = True

        if fail:
            raise ValueError(
                f"Invalid Run name: {value} the Run name must be lower case "
                "alphanumeric characters or '-', start with an alphabetic "
                "character, and end with an alphanumeric character (e.g. 'my-name', "
                " or 'abc-123')."
            )

    return value


def _aws_node_to_nickname():
    aws_node_to_nicknames = OrderedDict({
        'p3.16xlarge': '8_v100_16gb',
        'p3dn.24xlarge': '8_v100_32gb',
        'g4dn.metal': '8_t4_16gb',
        'p2.8xlarge': '8_k80_12gb',
        'p3.8xlarge': '4_v100_16gb',
        'g4dn.12xlarge': '4_t4_16gb',
        'g3.16xlarge': '4_m60_8gb',
        'g3.8xlarge': '2_m60_8gb',
        'p3.2xlarge': '1_v100_16gb',
        # 'p4d.24xlarge': '8_a100_40gb',  # currently not supported
        'g4dn.8xlarge': '1_t4_16gb',
        'g4dn.4xlarge': '1_t4_16gb',
        'g4dn.2xlarge': '1_t4_16gb',
        'g4dn.xlarge': '1_t4_16gb',
        'g4dn.16xlarge': '1_t4_16gb',
        'p2.xlarge': '1_k80_12gb',
        'g3s.xlarge': '1_m60_8gb',
        'g3.4xlarge': '1_m60_8gb',
        't2.large': '2_cpu_8gb',
        't2.medium': '2_cpu_4gb'
    })
    return aws_node_to_nicknames


def _nickname_to_aws_nodes():
    aws_node_to_nickname = _aws_node_to_nickname()
    aws_nickname_to_node = {v: k for k, v in aws_node_to_nickname.items()}
    return aws_nickname_to_node


def resolve_instance_type_nickname(value):
    """
    Enables instance type shortcuts like:
    2_cpu_4gb for t2.large
    """
    nickname = value.lower()

    aws_nickname_to_node = _nickname_to_aws_nodes()

    # validate potential options for the node name
    possible_values = list(aws_nickname_to_node.keys()) + list(aws_nickname_to_node.values())
    if nickname not in possible_values:
        possible_options = '\n'.join(list(aws_nickname_to_node.keys()))
        ValueError(f'{nickname} is not an available instance_type\n try one of these:\n{possible_options}')

    instance_type = nickname

    # if the value has _ then it's a nickname
    if '_' in nickname:
        instance_type = aws_nickname_to_node[nickname]
    return instance_type


def _get_instance_types(ctx, args, incomplete):
    # TODO: these should be retrieved from backend
    return list(_aws_node_to_nickname().keys())


def read_config(path: Union[str, Path]) -> dict:
    """
    Parameters
    ----------
    path:
        path to file on disk
    """
    if not isinstance(path, Path):
        path = Path(path)
    if not path.is_file():
        raise FileNotFoundError(f'grid config file {path} does not exist')

    #  Loads the YML file as passed by the user.
    try:
        with open(path) as f:
            grid_config = yaml.safe_load(f.read())
        if not isinstance(grid_config, dict):
            raise Exception("Unexpected file structure")
    except Exception as e:
        raise ValueError(f'Could not load your YAML config file: {e}')

    #  Adds required structure to the base YML file, if that structure isn't there.
    if 'compute' not in grid_config:
        grid_config['compute'] = {}
    if 'train' not in grid_config['compute']:
        grid_config['compute']['train'] = {}
    return grid_config


def check_description_isnt_too_long(value: str):
    """Click callback that checks if the description isn't too long."""
    if value is not None and len(value) > 200:
        raise ValueError('Description should have at most ' f'200 characters, yours has {len(value)}.')
    return value



from typing import Dict, List, Optional

from grid.openapi import (
    Body1,
    Externalv1Session,
    V1CreateSessionRequest,
    V1CreateSessionResponse,
    V1DatastoreInput,
    V1DeleteSessionResponse,
    V1GetSessionResponse,
    V1ListSessionsResponse,
    V1Resources,
    V1SessionSpec,
    V1SessionState,
    V1UpdateSessionResponse,
)
from grid.sdk.datastores import Datastore
from grid.sdk.rest.client import GridRestClient
from grid.sdk.rest.datastores import datastore_dsn_from_id
from grid.sdk.rest.exceptions import throw_with_message


@throw_with_message
def get_session_from_name(c: GridRestClient, name: str, cluster_id: str) -> Externalv1Session:
    """Find the id of a session with some name on a cluster.

    Parameters
    ----------
    c
        Client
    name
        the name of the session to find the ID of
    cluster_id
        which cluster should be used to find the datastore in.

    Returns
    -------
    str
       The ID of the session.
    """
    sessions: V1ListSessionsResponse = c.session_service_list_sessions(
        cluster_id=cluster_id, phase_in=[V1SessionState.RUNNING, V1SessionState.PAUSED, V1SessionState.PENDING]
    )
    for session in sessions.sessions:
        session: Externalv1Session
        if session.name == name:
            return session

    raise KeyError(f"could not find session with name {name} in cluster {cluster_id}")


@throw_with_message
def list_sessions(c: GridRestClient,
                  cluster_id: Optional[str] = None,
                  user_ids: Optional[List[str]] = None) -> List[Externalv1Session]:
    kwargs = {}
    if cluster_id is not None:
        kwargs['cluster_id'] = cluster_id
    if user_ids is not None:
        kwargs['user_ids'] = user_ids
    kwargs['phase_not_in'] = [V1SessionState.DELETED]

    resp: V1ListSessionsResponse = c.session_service_list_sessions(**kwargs)
    return resp.sessions


@throw_with_message
def get_session_from_id(c: GridRestClient, session_id: str) -> V1GetSessionResponse:
    return c.session_service_get_session(id=session_id)


@throw_with_message
def create_session(
    c: GridRestClient,
    name: str,
    instance_type: str,
    cluster_id: Optional[str] = None,
    datastores: Optional[Dict[str, str]] = None,  # datastore_id: mount_path
    disk_size_gb: str = '200',
    use_spot: bool = False
) -> V1CreateSessionResponse:
    """Create a session on the grid platform.

    Parameters
    ----------
    c
        API client
    name
        human readable name of the session which is created
    instance_type
        cloud provider instance type used to start the session on.
    cluster_id
        cluster ID to start the session on
    disk_size_gb
        The size of the disk to use.
    datastores
        mapping of datastore id to datastore mount dir
    use_spot
        if we should use a spot instance or not.

    Returns
    -------
    str
        ID of the session which was created.
    """
    datastore_mounts = None
    if datastores:
        datastore_mounts = [
            V1DatastoreInput(dsn=datastore_dsn_from_id(id=key), mount_path=value) for key, value in datastores.items()
        ]

    request = V1CreateSessionRequest(
        name=name,
        spec=V1SessionSpec(
            cluster_id=cluster_id,
            use_spot=use_spot,
            desired_state=V1SessionState.RUNNING,
            instance_type=instance_type,
            datastores=datastore_mounts,
            resources=V1Resources(storage_gb=disk_size_gb, )
        )
    )
    resp = c.session_service_create_session(request)
    return resp


@throw_with_message
def pause_session(c: GridRestClient, session_id: str, cluster_id: Optional[str] = None) -> V1UpdateSessionResponse:
    """Pause a session with a given name.

    Parameters
    ----------
    c
        Open API client
    session_id
        session id
    cluster_id
        cluster containing the session which will be paused

    Returns
    -------
    str
        ID of the session which is paused.
    """
    session = get_session_from_id(c=c, session_id=session_id)
    session.spec.desired_state = V1SessionState.PAUSED
    resp: V1UpdateSessionResponse = c.session_service_update_session(
        id=session.id, body=Body1(name=session.name, spec=session.spec)
    )
    return resp


@throw_with_message
def delete_session(c: GridRestClient, session_id: str, cluster_id: Optional[str] = None) -> V1DeleteSessionResponse:
    """Delete a session with a given name.

    Parameters
    ----------
    c
        Open API client
    session_id
        id of the session to delete
    cluster_id
        cluster containing the session which will be deleted

    Returns
    -------
    str
        ID of the session which is deleted.
    """
    return c.session_service_delete_session(id=session_id)


@throw_with_message
def resume_session(c: GridRestClient, session_id: str, cluster_id: Optional[str] = None) -> V1UpdateSessionResponse:
    """Resumes a session with a given name.

    Parameters
    ----------
    c
        Open API client
    session_id
        id of the session to pause
    cluster_id
        cluster containing the session which will be resumed

    Returns
    -------
    V1UpdateSessionResponse
        session which is resumed
    """
    session = get_session_from_id(c, session_id=session_id)

    spec: V1SessionSpec = session.spec
    spec.desired_state = V1SessionState.RUNNING
    session.spec = spec
    resp: V1UpdateSessionResponse = c.session_service_update_session(
        id=session.id, body=Body1(name=session.name, spec=spec)
    )
    return resp


@throw_with_message
def change_session_instance_type(
    c: GridRestClient,
    session_id: str,
    instance_type,
    use_spot: Optional[bool] = None,
    cluster_id: Optional[str] = None
) -> V1UpdateSessionResponse:
    """Change the instance type of a session.

    Parameters
    ----------
    c
        Open API Client
    session_id
        name of the session to alter
    instance_type
        new instance type to use for the session
    use_spot
        if true, use interuptable spot instances (which come at a steap discount,
        but which can be interrupted and shut down at any point in time depending
        on cloud provider instance type demand). If false, use an on-demand instance.

        By default this value is None, indicating that no change will be made to the
        current configuratoin.
    cluster_id
        cluster ID of the session to change the instance type of

    Returns
    -------
    V1UpdateSessionResponse
        response from the server
    """
    session = get_session_from_id(c, session_id=session_id)
    session.spec.instance_type = instance_type
    if use_spot is not None:
        session.spec.use_spot = use_spot

    result = c.session_service_update_session(
        id=session.id,
        body=Body1(name=session.name, spec=session.spec),
    )
    return result



import time
from typing import List

import click
import yaspin

from grid import Experiment, Run
from grid.cli import rich_click
from grid.sdk import env
from grid.sdk.rest.client import GridRestClient
from grid.sdk.client import create_swagger_client
from grid.sdk.rest.exceptions import GridException
from grid.openapi.api import cluster_service_api
from grid.openapi import V1ClusterState

CLUSTER_STATE_CHECKING_TIMEOUT = 60
MAX_CLUSTER_WAIT_TIME = 5400


@rich_click.group()
def delete() -> None:
    """Allows you to delete grid resources."""
    pass


def doublecheck(item: str):
    warning_str = click.style('WARNING!', fg='red')
    message = f"""

    {warning_str}

    Your are about to delete the {item}.
    This will delete all the associated artifacts, logs, and metadata.

    Are you sure you want to do this?

   """
    click.confirm(message, abort=True)


@delete.command()
@rich_click.argument('experiment_names', type=str, required=True, nargs=-1, help='Experiment names to delete.')
def experiment(experiment_names: List[str]):
    """Delete some set of EXPERIMENT_NAMES from grid.

    This process is immediate and irreversible, deletion permanently removes not only
    the record of the experiment, but all associated artifacts, metrics, logs, etc.
    """
    doublecheck(str(experiment_names))
    for name in experiment_names:
        spinner = yaspin.yaspin(text=f'Stopping experiment {name}...', color="yellow")
        spinner.start()
        try:
            exp = Experiment(name=name)
            if not exp.exists:
                raise RuntimeError(
                    f"Experiment {name} does not exist in the cluster {env.CONTEXT}. "
                    f"If you are trying to delete an experiment in another cluster, "
                    f"try setting the default cluster with "
                    f"`grid user set-default-cluster <cluster_name>` first"
                )
            exp.delete()
            spinner.ok("✔")
            styled_name = click.style(name, fg='blue')
            click.echo(f'Experiment {styled_name} was deleted successfully.')
        except GridException as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(str(e))
        except Exception as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(f"Deletion failed for experiment {name}: {e}")
        finally:
            spinner.stop()


@delete.command()
@rich_click.argument('run_names', type=str, required=True, nargs=-1, help='Run names to delete.')
def run(run_names: List[str]):
    """Delete some set of RUN_NAMES from grid.

    Deleting a run also deletes all experiments contained within the run.

    This process is immediate and irreversible, deletion permanently removes not only
    the record of the run, but all associated experiments, artifacts, metrics, logs, etc.
    """
    doublecheck(str(run_names))
    for name in run_names:
        spinner = yaspin.yaspin(text=f'Stopping run {name}...', color="yellow")
        spinner.start()
        try:
            run_obj = Run(name=name)
            if not run_obj.exists:
                raise RuntimeError(
                    f"Run {name} does not exist in the cluster {env.CONTEXT}. "
                    f"If you are trying to delete a run in another cluster, "
                    f"try setting the default cluster with "
                    f"`grid user set-default-cluster <cluster_name>` first"
                )
            run_obj.delete()
            spinner.ok("✔")
            styled_name = click.style(name, fg='blue')
            click.echo(f'Run {styled_name} was deleted successfully.')
        except GridException as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(str(e))
        except Exception as e:
            if spinner:
                spinner.fail("✘")
            raise click.ClickException(f"Deletion failed for run {name}: {e}")
        finally:
            spinner.stop()


@delete.command()
@rich_click.argument('cluster', type=str, help='Cluster id to delete.')
@click.option(
    '--force',
    'force',
    type=bool,
    required=False,
    default=False,
    is_flag=True,
    help='Force delete cluster from grid system. This does NOT delete any resources created by the cluster, '
    'just cleaning up the entry from the grid system. You should not use this under normal circumstances',
)
@click.option(
    '--wait',
    'wait',
    type=bool,
    required=False,
    default=False,
    is_flag=True,
    help='using this flag CLI will wait until the cluster is deleted',
)
def cluster(cluster: str, force: bool = False, wait: bool = False):
    """Delete CLUSTER and all associated AWS resources.

    Deleting a run also deletes all Runs and Experiments which were started
    on the cluster. deletion permanently removes not only the record of all
    runs on a cluster, but all associated experiments, artifacts, metrics, logs, etc.

    This process may take a few minutes to complete, but once started is irriversable.
    Deletion permanently removes not only cluster from being managed by grid, but tears
    down every resource grid managed (for that cluster id) in the host cloud. All object
    stores, container registries, logs, compute nodes, volumes, etc. are deleted and
    cannot be recovered.
    """
    if force:
        click.echo(
            "Force deleting cluster. This will cause grid to forget "
            "about the cluster and any experiments, sessions, datastores, "
            "tensorboards and other resources running on it.\n"
            "WARNING: this will not clean up any resources managed by grid\n"
            "Check your cloud provider that any existing cloud resources are deleted"
        )
        click.confirm('Do you want to continue?', abort=True)

    client = GridRestClient(api_client=create_swagger_client())
    client.cluster_service_delete_cluster(id=cluster, force=force)
    click.echo("Cluster deletion triggered successfully")

    if wait:
        start = time.time()
        elapsed = 0
        api_client = create_swagger_client()
        cluster_api = cluster_service_api.ClusterServiceApi(api_client)
        while elapsed < MAX_CLUSTER_WAIT_TIME:
            cluster_resp = cluster_api.cluster_service_list_clusters()
            cluster_to_del = None
            for clust in cluster_resp.clusters:
                if clust.id == cluster:
                    cluster_to_del = clust
                    break
            if cluster_to_del is not None:
                if cluster_to_del.status.phase == V1ClusterState.DELETED:
                    break
                elif cluster_to_del.status.phase == V1ClusterState.FAILED:
                    raise click.ClickException(f"Deletion failed for cluster {cluster}")
                time.sleep(CLUSTER_STATE_CHECKING_TIMEOUT)
            else:
                break
            elapsed = time.time() - start
        else:
            raise click.ClickException(f"Max time for cluster deletion is elapsed")



import json
import os
from pathlib import Path
from typing import Optional

from grid.sdk.auth import Credentials
from grid.sdk import env
from grid.sdk.api import list_clusters
from grid.sdk.user import set_default_cluster
from grid.sdk.rest import GridRestClient
from grid.openapi.models import V1LoginRequest
from grid.openapi import ApiClient, Configuration
from grid.sdk.auth import Credentials


def login(
    username: Optional[str] = None,
    user_id: Optional[str] = None,
    api_key: Optional[str] = None,
    *,
    _url=env.GRID_URL  # noqa - For testing purposes only. not user facing.
):
    """
    Log in with your grid.ai credentials for usage of the SDK in the running process.

    All parameters are optional. Calling ``login()`` without parameters will check if
    the ``GRID_USER_ID`` and ``GRID_API_KEY`` env vars have been set (using those if
    available), otherwise it will check for the file ``credentials.json`` in the
    machines ``$HOME/.grid`` directory (if it exists).

    If no credentials have been stored, then you must pass in your API key and either
    your username or user id (if you know it). Your user id and API key can be found
    by navigating to https://platform.grid.ai/#/settings?tabId=apikey.

    Parameters
    ----------
    username
        your grid username. This is either be your github username or email address,
        depending on what you use when signing into the grid platform at:
        https://platform.grid.ai

    user_id
        Your grid user id. This can be found by visiting https://platform.grid.ai/#/settings?tabId=apikey.
    api_key
        Your grid API key. This can be found by visiting https://platform.grid.ai/#/settings?tabId=apikey.
    """
    if user_id and api_key:
        creds = Credentials(user_id=user_id, api_key=api_key)
        _create_credentials_file(creds)
    elif username and api_key:
        # create an unauthenticated client
        configuration = Configuration()
        configuration.host = _url
        configuration.ssl_ca_cert = env.SSL_CA_CERT
        client = GridRestClient(ApiClient(configuration=configuration))

        # get authentication token
        token_resp = client.auth_service_login(V1LoginRequest(
            username=username,
            api_key=api_key,
        ))
        client.api_client.set_default_header("Authorization", f"Bearer {token_resp.token}")

        # now that we're authenticated, get the user ID
        user_resp = client.auth_service_get_user()
        creds = Credentials(user_id=user_resp.id, api_key=api_key)
        _create_credentials_file(creds)

    elif (username or user_id) or api_key:
        raise ValueError("Either (`user_id` OR `username`) AND `api_key` need to be set, or none set.")
    else:
        creds = Credentials.from_locale()

    # set the GRID_URL before new api clients are created
    os.environ['GRID_URL'] = _url
    env.reset_global_variables()

    # adding context to the user settings
    # this will overwrite the context user set - when we have the set context command.
    # solution to that problem is to have DEFAULT_CONTEXT and SELECTED_CONTEXT
    cluster_list = list_clusters(is_global=True)
    default_cluster = cluster_list.default_cluster
    set_default_cluster(cluster_id=default_cluster)

    settings_path = Path(env.GRID_SETTINGS_PATH)
    user_settings = json.load(settings_path.open())
    user_settings['grid_url'] = _url
    with settings_path.open('w') as file:
        json.dump(user_settings, file, ensure_ascii=False, indent=4)

    # Set the user credentials in the os env
    os.environ['GRID_USER_ID'] = creds.user_id
    os.environ['GRID_API_KEY'] = creds.api_key
    creds = Credentials.from_locale()
    return True


def _create_credentials_file(creds: Credentials):
    Path(env.GRID_CREDENTIAL_PATH).parent.mkdir(parents=True, exist_ok=True)
    with Path(env.GRID_CREDENTIAL_PATH).open('w') as file:
        json.dump({'UserID': creds.user_id, 'APIKey': creds.api_key}, file, ensure_ascii=False, indent=4)



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1ClusterSpec(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'cluster_type': 'V1ClusterType',
        'cost_factor': 'str',
        'desired_state': 'V1ClusterState',
        'driver': 'V1ClusterDriver',
        'gridlet_version': 'str',
        'padding_factor': 'str',
        'pause_automation': 'bool',
        'performance_profile': 'V1ClusterPerformanceProfile',
        'user_id': 'str'
    }

    attribute_map = {
        'cluster_type': 'clusterType',
        'cost_factor': 'costFactor',
        'desired_state': 'desiredState',
        'driver': 'driver',
        'gridlet_version': 'gridletVersion',
        'padding_factor': 'paddingFactor',
        'pause_automation': 'pauseAutomation',
        'performance_profile': 'performanceProfile',
        'user_id': 'userId'
    }

    def __init__(self, cluster_type: 'V1ClusterType' = None, cost_factor: 'str' = None, desired_state: 'V1ClusterState' = None, driver: 'V1ClusterDriver' = None, gridlet_version: 'str' = None, padding_factor: 'str' = None, pause_automation: 'bool' = None, performance_profile: 'V1ClusterPerformanceProfile' = None, user_id: 'str' = None, _configuration=None):  # noqa: E501
        """V1ClusterSpec - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._cluster_type = None
        self._cost_factor = None
        self._desired_state = None
        self._driver = None
        self._gridlet_version = None
        self._padding_factor = None
        self._pause_automation = None
        self._performance_profile = None
        self._user_id = None
        self.discriminator = None

        if cluster_type is not None:
            self.cluster_type = cluster_type
        if cost_factor is not None:
            self.cost_factor = cost_factor
        if desired_state is not None:
            self.desired_state = desired_state
        if driver is not None:
            self.driver = driver
        if gridlet_version is not None:
            self.gridlet_version = gridlet_version
        if padding_factor is not None:
            self.padding_factor = padding_factor
        if pause_automation is not None:
            self.pause_automation = pause_automation
        if performance_profile is not None:
            self.performance_profile = performance_profile
        if user_id is not None:
            self.user_id = user_id

    @property
    def cluster_type(self) -> 'V1ClusterType':
        """Gets the cluster_type of this V1ClusterSpec.  # noqa: E501


        :return: The cluster_type of this V1ClusterSpec.  # noqa: E501
        :rtype: V1ClusterType
        """
        return self._cluster_type

    @cluster_type.setter
    def cluster_type(self, cluster_type: 'V1ClusterType'):
        """Sets the cluster_type of this V1ClusterSpec.


        :param cluster_type: The cluster_type of this V1ClusterSpec.  # noqa: E501
        :type: V1ClusterType
        """

        self._cluster_type = cluster_type

    @property
    def cost_factor(self) -> 'str':
        """Gets the cost_factor of this V1ClusterSpec.  # noqa: E501

        costController needs for calculating costOfGoods - string since it's decimal. Note: This value is stored in a separate column in the database, not with the rest of the cluster JSON in spec.  # noqa: E501

        :return: The cost_factor of this V1ClusterSpec.  # noqa: E501
        :rtype: str
        """
        return self._cost_factor

    @cost_factor.setter
    def cost_factor(self, cost_factor: 'str'):
        """Sets the cost_factor of this V1ClusterSpec.

        costController needs for calculating costOfGoods - string since it's decimal. Note: This value is stored in a separate column in the database, not with the rest of the cluster JSON in spec.  # noqa: E501

        :param cost_factor: The cost_factor of this V1ClusterSpec.  # noqa: E501
        :type: str
        """

        self._cost_factor = cost_factor

    @property
    def desired_state(self) -> 'V1ClusterState':
        """Gets the desired_state of this V1ClusterSpec.  # noqa: E501


        :return: The desired_state of this V1ClusterSpec.  # noqa: E501
        :rtype: V1ClusterState
        """
        return self._desired_state

    @desired_state.setter
    def desired_state(self, desired_state: 'V1ClusterState'):
        """Sets the desired_state of this V1ClusterSpec.


        :param desired_state: The desired_state of this V1ClusterSpec.  # noqa: E501
        :type: V1ClusterState
        """

        self._desired_state = desired_state

    @property
    def driver(self) -> 'V1ClusterDriver':
        """Gets the driver of this V1ClusterSpec.  # noqa: E501


        :return: The driver of this V1ClusterSpec.  # noqa: E501
        :rtype: V1ClusterDriver
        """
        return self._driver

    @driver.setter
    def driver(self, driver: 'V1ClusterDriver'):
        """Sets the driver of this V1ClusterSpec.


        :param driver: The driver of this V1ClusterSpec.  # noqa: E501
        :type: V1ClusterDriver
        """

        self._driver = driver

    @property
    def gridlet_version(self) -> 'str':
        """Gets the gridlet_version of this V1ClusterSpec.  # noqa: E501


        :return: The gridlet_version of this V1ClusterSpec.  # noqa: E501
        :rtype: str
        """
        return self._gridlet_version

    @gridlet_version.setter
    def gridlet_version(self, gridlet_version: 'str'):
        """Sets the gridlet_version of this V1ClusterSpec.


        :param gridlet_version: The gridlet_version of this V1ClusterSpec.  # noqa: E501
        :type: str
        """

        self._gridlet_version = gridlet_version

    @property
    def padding_factor(self) -> 'str':
        """Gets the padding_factor of this V1ClusterSpec.  # noqa: E501

        costController needs for calculating paddingValue - string since it's decimal Note: This value is stored in a separate column in the database, not with the rest of the cluster JSON in spec.  # noqa: E501

        :return: The padding_factor of this V1ClusterSpec.  # noqa: E501
        :rtype: str
        """
        return self._padding_factor

    @padding_factor.setter
    def padding_factor(self, padding_factor: 'str'):
        """Sets the padding_factor of this V1ClusterSpec.

        costController needs for calculating paddingValue - string since it's decimal Note: This value is stored in a separate column in the database, not with the rest of the cluster JSON in spec.  # noqa: E501

        :param padding_factor: The padding_factor of this V1ClusterSpec.  # noqa: E501
        :type: str
        """

        self._padding_factor = padding_factor

    @property
    def pause_automation(self) -> 'bool':
        """Gets the pause_automation of this V1ClusterSpec.  # noqa: E501


        :return: The pause_automation of this V1ClusterSpec.  # noqa: E501
        :rtype: bool
        """
        return self._pause_automation

    @pause_automation.setter
    def pause_automation(self, pause_automation: 'bool'):
        """Sets the pause_automation of this V1ClusterSpec.


        :param pause_automation: The pause_automation of this V1ClusterSpec.  # noqa: E501
        :type: bool
        """

        self._pause_automation = pause_automation

    @property
    def performance_profile(self) -> 'V1ClusterPerformanceProfile':
        """Gets the performance_profile of this V1ClusterSpec.  # noqa: E501


        :return: The performance_profile of this V1ClusterSpec.  # noqa: E501
        :rtype: V1ClusterPerformanceProfile
        """
        return self._performance_profile

    @performance_profile.setter
    def performance_profile(self, performance_profile: 'V1ClusterPerformanceProfile'):
        """Sets the performance_profile of this V1ClusterSpec.


        :param performance_profile: The performance_profile of this V1ClusterSpec.  # noqa: E501
        :type: V1ClusterPerformanceProfile
        """

        self._performance_profile = performance_profile

    @property
    def user_id(self) -> 'str':
        """Gets the user_id of this V1ClusterSpec.  # noqa: E501


        :return: The user_id of this V1ClusterSpec.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id: 'str'):
        """Sets the user_id of this V1ClusterSpec.


        :param user_id: The user_id of this V1ClusterSpec.  # noqa: E501
        :type: str
        """

        self._user_id = user_id

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1ClusterSpec, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1ClusterSpec') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1ClusterSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1ClusterSpec') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1ClusterSpec):
            return True

        return self.to_dict() != other.to_dict()



import urllib3

from grid.cli.utilities import get_graphql_url
from grid.metadata import __version__
from grid.openapi import ApiClient, Configuration
from grid.sdk.auth import Credentials
from grid.sdk.client.grid_gql import gql_client
from grid.websocket.client import WebSocketClient
import grid.sdk.env as env


def create_gql_client(*, websocket: bool = False):
    # since we are manipulating the GRID_URL on the fly, we always need to check the env variable
    url = get_graphql_url(env.GRID_URL)
    creds = Credentials.from_locale()
    res = gql_client(url=url, creds=creds, websocket=websocket)
    return res


def create_websocket_client():
    if not env.CONTEXT:
        raise RuntimeError('Default cluster is not found. Try logging in again!')
    creds = Credentials.from_locale()
    token = urllib3.util.make_headers(basic_auth=creds.user_id + ':' + creds.api_key).get('authorization')
    return WebSocketClient(base_url=env.GRID_URL, token=token)


def create_swagger_client(check_context=True):
    """
    Create the swagger client to use the autogenerated code

    Parameters
    ----------
    check_context: bool
        If true, check if the context is set. It's only false for APIs that doesn't
        need the context information i.e login
    """
    if check_context and not env.CONTEXT:
        raise RuntimeError('Default cluster is not found. Try logging in again!')
    url = env.GRID_URL
    creds = Credentials.from_locale()
    configuration = Configuration()
    configuration.host = url
    configuration.username = creds.user_id
    configuration.password = creds.api_key
    # for custom certs we need to hint urllib to use one of them if available (requests package would use any of them). these two are also use
    # during artifacts to actually reverted and not use the custom certs if present in /grid-cli/grid/cli/cli/artifacts.py
    configuration.ssl_ca_cert = env.SSL_CA_CERT

    # TODO Increase the pool_thread (and make it configurable) and use a singleton
    api_client = ApiClient(
        configuration, header_name="Authorization", header_value=configuration.get_basic_auth_token()
    )
    api_client.user_agent = f'Grid-CLI-{__version__}'
    return api_client



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1EKSCusterDriverStatus(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'bastion_ami': 'str',
        'builder_ami': 'str',
        'cpu_ami': 'str',
        'ecr_registry_id': 'str',
        'ecr_repository_uri': 'str',
        'fsx_global_role_arn': 'str',
        'fsx_sg_id': 'str',
        'gpu_ami': 'str',
        'gridlet_role_arn': 'str',
        'user_workload_role_arn': 'str',
        'vpc_id': 'str',
        'vpc_private_subnet_ids': 'list[str]',
        'vpc_public_subnet_ids': 'list[str]'
    }

    attribute_map = {
        'bastion_ami': 'bastionAmi',
        'builder_ami': 'builderAmi',
        'cpu_ami': 'cpuAmi',
        'ecr_registry_id': 'ecrRegistryId',
        'ecr_repository_uri': 'ecrRepositoryUri',
        'fsx_global_role_arn': 'fsxGlobalRoleArn',
        'fsx_sg_id': 'fsxSgId',
        'gpu_ami': 'gpuAmi',
        'gridlet_role_arn': 'gridletRoleArn',
        'user_workload_role_arn': 'userWorkloadRoleArn',
        'vpc_id': 'vpcId',
        'vpc_private_subnet_ids': 'vpcPrivateSubnetIds',
        'vpc_public_subnet_ids': 'vpcPublicSubnetIds'
    }

    def __init__(self, bastion_ami: 'str' = None, builder_ami: 'str' = None, cpu_ami: 'str' = None, ecr_registry_id: 'str' = None, ecr_repository_uri: 'str' = None, fsx_global_role_arn: 'str' = None, fsx_sg_id: 'str' = None, gpu_ami: 'str' = None, gridlet_role_arn: 'str' = None, user_workload_role_arn: 'str' = None, vpc_id: 'str' = None, vpc_private_subnet_ids: 'list[str]' = None, vpc_public_subnet_ids: 'list[str]' = None, _configuration=None):  # noqa: E501
        """V1EKSCusterDriverStatus - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._bastion_ami = None
        self._builder_ami = None
        self._cpu_ami = None
        self._ecr_registry_id = None
        self._ecr_repository_uri = None
        self._fsx_global_role_arn = None
        self._fsx_sg_id = None
        self._gpu_ami = None
        self._gridlet_role_arn = None
        self._user_workload_role_arn = None
        self._vpc_id = None
        self._vpc_private_subnet_ids = None
        self._vpc_public_subnet_ids = None
        self.discriminator = None

        if bastion_ami is not None:
            self.bastion_ami = bastion_ami
        if builder_ami is not None:
            self.builder_ami = builder_ami
        if cpu_ami is not None:
            self.cpu_ami = cpu_ami
        if ecr_registry_id is not None:
            self.ecr_registry_id = ecr_registry_id
        if ecr_repository_uri is not None:
            self.ecr_repository_uri = ecr_repository_uri
        if fsx_global_role_arn is not None:
            self.fsx_global_role_arn = fsx_global_role_arn
        if fsx_sg_id is not None:
            self.fsx_sg_id = fsx_sg_id
        if gpu_ami is not None:
            self.gpu_ami = gpu_ami
        if gridlet_role_arn is not None:
            self.gridlet_role_arn = gridlet_role_arn
        if user_workload_role_arn is not None:
            self.user_workload_role_arn = user_workload_role_arn
        if vpc_id is not None:
            self.vpc_id = vpc_id
        if vpc_private_subnet_ids is not None:
            self.vpc_private_subnet_ids = vpc_private_subnet_ids
        if vpc_public_subnet_ids is not None:
            self.vpc_public_subnet_ids = vpc_public_subnet_ids

    @property
    def bastion_ami(self) -> 'str':
        """Gets the bastion_ami of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The bastion_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._bastion_ami

    @bastion_ami.setter
    def bastion_ami(self, bastion_ami: 'str'):
        """Sets the bastion_ami of this V1EKSCusterDriverStatus.


        :param bastion_ami: The bastion_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._bastion_ami = bastion_ami

    @property
    def builder_ami(self) -> 'str':
        """Gets the builder_ami of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The builder_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._builder_ami

    @builder_ami.setter
    def builder_ami(self, builder_ami: 'str'):
        """Sets the builder_ami of this V1EKSCusterDriverStatus.


        :param builder_ami: The builder_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._builder_ami = builder_ami

    @property
    def cpu_ami(self) -> 'str':
        """Gets the cpu_ami of this V1EKSCusterDriverStatus.  # noqa: E501

        AMI to be used for CPU nodes.  # noqa: E501

        :return: The cpu_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._cpu_ami

    @cpu_ami.setter
    def cpu_ami(self, cpu_ami: 'str'):
        """Sets the cpu_ami of this V1EKSCusterDriverStatus.

        AMI to be used for CPU nodes.  # noqa: E501

        :param cpu_ami: The cpu_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._cpu_ami = cpu_ami

    @property
    def ecr_registry_id(self) -> 'str':
        """Gets the ecr_registry_id of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The ecr_registry_id of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._ecr_registry_id

    @ecr_registry_id.setter
    def ecr_registry_id(self, ecr_registry_id: 'str'):
        """Sets the ecr_registry_id of this V1EKSCusterDriverStatus.


        :param ecr_registry_id: The ecr_registry_id of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._ecr_registry_id = ecr_registry_id

    @property
    def ecr_repository_uri(self) -> 'str':
        """Gets the ecr_repository_uri of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The ecr_repository_uri of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._ecr_repository_uri

    @ecr_repository_uri.setter
    def ecr_repository_uri(self, ecr_repository_uri: 'str'):
        """Sets the ecr_repository_uri of this V1EKSCusterDriverStatus.


        :param ecr_repository_uri: The ecr_repository_uri of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._ecr_repository_uri = ecr_repository_uri

    @property
    def fsx_global_role_arn(self) -> 'str':
        """Gets the fsx_global_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The fsx_global_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._fsx_global_role_arn

    @fsx_global_role_arn.setter
    def fsx_global_role_arn(self, fsx_global_role_arn: 'str'):
        """Sets the fsx_global_role_arn of this V1EKSCusterDriverStatus.


        :param fsx_global_role_arn: The fsx_global_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._fsx_global_role_arn = fsx_global_role_arn

    @property
    def fsx_sg_id(self) -> 'str':
        """Gets the fsx_sg_id of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The fsx_sg_id of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._fsx_sg_id

    @fsx_sg_id.setter
    def fsx_sg_id(self, fsx_sg_id: 'str'):
        """Sets the fsx_sg_id of this V1EKSCusterDriverStatus.


        :param fsx_sg_id: The fsx_sg_id of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._fsx_sg_id = fsx_sg_id

    @property
    def gpu_ami(self) -> 'str':
        """Gets the gpu_ami of this V1EKSCusterDriverStatus.  # noqa: E501

        AMI to be used for GPU nodes.  # noqa: E501

        :return: The gpu_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._gpu_ami

    @gpu_ami.setter
    def gpu_ami(self, gpu_ami: 'str'):
        """Sets the gpu_ami of this V1EKSCusterDriverStatus.

        AMI to be used for GPU nodes.  # noqa: E501

        :param gpu_ami: The gpu_ami of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._gpu_ami = gpu_ami

    @property
    def gridlet_role_arn(self) -> 'str':
        """Gets the gridlet_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The gridlet_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._gridlet_role_arn

    @gridlet_role_arn.setter
    def gridlet_role_arn(self, gridlet_role_arn: 'str'):
        """Sets the gridlet_role_arn of this V1EKSCusterDriverStatus.


        :param gridlet_role_arn: The gridlet_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._gridlet_role_arn = gridlet_role_arn

    @property
    def user_workload_role_arn(self) -> 'str':
        """Gets the user_workload_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The user_workload_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._user_workload_role_arn

    @user_workload_role_arn.setter
    def user_workload_role_arn(self, user_workload_role_arn: 'str'):
        """Sets the user_workload_role_arn of this V1EKSCusterDriverStatus.


        :param user_workload_role_arn: The user_workload_role_arn of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._user_workload_role_arn = user_workload_role_arn

    @property
    def vpc_id(self) -> 'str':
        """Gets the vpc_id of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The vpc_id of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: str
        """
        return self._vpc_id

    @vpc_id.setter
    def vpc_id(self, vpc_id: 'str'):
        """Sets the vpc_id of this V1EKSCusterDriverStatus.


        :param vpc_id: The vpc_id of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: str
        """

        self._vpc_id = vpc_id

    @property
    def vpc_private_subnet_ids(self) -> 'list[str]':
        """Gets the vpc_private_subnet_ids of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The vpc_private_subnet_ids of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: list[str]
        """
        return self._vpc_private_subnet_ids

    @vpc_private_subnet_ids.setter
    def vpc_private_subnet_ids(self, vpc_private_subnet_ids: 'list[str]'):
        """Sets the vpc_private_subnet_ids of this V1EKSCusterDriverStatus.


        :param vpc_private_subnet_ids: The vpc_private_subnet_ids of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: list[str]
        """

        self._vpc_private_subnet_ids = vpc_private_subnet_ids

    @property
    def vpc_public_subnet_ids(self) -> 'list[str]':
        """Gets the vpc_public_subnet_ids of this V1EKSCusterDriverStatus.  # noqa: E501


        :return: The vpc_public_subnet_ids of this V1EKSCusterDriverStatus.  # noqa: E501
        :rtype: list[str]
        """
        return self._vpc_public_subnet_ids

    @vpc_public_subnet_ids.setter
    def vpc_public_subnet_ids(self, vpc_public_subnet_ids: 'list[str]'):
        """Sets the vpc_public_subnet_ids of this V1EKSCusterDriverStatus.


        :param vpc_public_subnet_ids: The vpc_public_subnet_ids of this V1EKSCusterDriverStatus.  # noqa: E501
        :type: list[str]
        """

        self._vpc_public_subnet_ids = vpc_public_subnet_ids

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1EKSCusterDriverStatus, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1EKSCusterDriverStatus') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1EKSCusterDriverStatus):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1EKSCusterDriverStatus') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1EKSCusterDriverStatus):
            return True

        return self.to_dict() != other.to_dict()



def user_not_authenticated() -> str:
    """Technically this catches either un-authenticated, or blocked users.

     This is used for both cases since we don't want users to know they have been blocked.
    """
    return (
        "You are either not logged into grid, or are not authorized to check "
        "the requested resource. Please login via the `grid login` command."
    )


def datastore_invalid_source(source: str) -> str:
    return (
        f'Unknown Source type for {source}. We accept local file paths, '
        f'HTTP/HTTPS URLs or Amazon S3 Buckets (s3://foo/bar/). Please '
        f'check your input and try again'
    )


def datastore_invalid_no_copy_option(source: str) -> str:
    return (
        f"The no-copy option was passed when creating a datastore from {source}. "
        "This option is only valid for s3 buckets. Please consider removing the "
        "flag or ensuring that your path starts with `s3://`"
    )


def datastore_local_source_dir_is_empty(source) -> str:
    return f"The directory {source} is empty. To create a datastore, atleast one file must exist."


def datastore_upload_process_cancelled() -> str:
    return (
        "You've cancelled your upload, but your progress has been saved. You can run: \n"
        "\n"
        "$ grid datastore resume\n"
        "\n"
        "to resume your upload."
    )


def datastore_s3_source_does_not_end_in_slash(source: str) -> str:
    return (
        f"The {source} S3 URL must end in a '/' in order to specify a directory. "
        f"Single-file S3 datastores are not currently supported. Please rerun your "
        f"last commandand add a '/' to the end of {source} so that it looks like: "
        f"{source}/"
    )


def datastore_invalid_fsx_option(source: str) -> str:
    return (
        f"The --hpd option was passed when creating a datastore from {source}. "
        "This option is only valid for s3 buckets. Please consider removing the "
        "flag or ensuring that your path starts with `s3://`"
    )


def datastore_invalid_fsx_throughput(throughput: int) -> str:
    return (
        f"The --hpd-throughput option was passed when creating a high-performance datastore "
        f"but the value ({throughput}) is invalid , please ensure this value is one of the following: "
        f"[125, 250, 500, 1000]."
    )


def datastore_invalid_fsx_capacity(capacity: int) -> str:
    return (
        f"The --hpd-capacity option was passed when creating a high-performance backed datastore "
        f"but the value ({capacity}) is invalid , please ensure this value is either 1200, "
        f"2400 or a multiple of 2400. "
    )



from typing import List, Optional, Dict

from grid.openapi.models import (
    Body8,
    V1CreateRunRequest,
    V1Run,
    V1RunSpec,
    V1RunState,
    V1RunActions,
    V1Resources,
    Externalv1ScratchSpace,
    Externalv1ImageSpec,
    V1DependencyFileInfo,
    V1PackageManager,
    V1DatastoreMounts,
    V1DeleteRunResponse,
)
from grid.sdk.rest.client import GridRestClient
from grid.sdk.rest.exceptions import throw_with_message


@throw_with_message
def create_run(
    client: GridRestClient,
    cluster_id: str,
    run_name: str,
    run_description: Optional[str],
    run_controller_command: str,
    source_code: str,
    localdir: bool,
    run_sweep_type: str,
    run_sweep_options: Optional[Dict[str, str]],
    scratch_mount_path: str,
    scratch_size_gb: int,
    instance_type: str,
    use_spot: bool,
    per_exp_resources_cpu: Optional[int] = None,
    per_exp_resources_gpu: Optional[int] = None,
    per_exp_resources_storage_gb: Optional[int] = None,
    per_exp_resources_extra: Optional[Dict[str, str]] = None,
    on_build_actions: Optional[List[str]] = None,
    on_build_start_actions: Optional[List[str]] = None,
    on_build_end_actions: Optional[List[str]] = None,
    on_experiment_start_actions: Optional[List[str]] = None,
    on_experiment_end_actions: Optional[List[str]] = None,
    image_dockerfile: Optional[str] = None,
    image_framework: Optional[str] = None,
    image_dep_file: Optional[str] = None,
    per_exp_env_vars: Optional[Dict[str, str]] = None,
    datastores: Optional[Dict[str, str]] = None,  # datastore_id: mount_path
    relative_work_dir: str = None,
    dry_run: bool = False,
    auto_resume: bool = False,
) -> V1Run:
    if image_dep_file and image_dep_file.endswith('.txt'):
        dep_file_info = V1DependencyFileInfo(
            path=image_dep_file,
            package_manager=V1PackageManager.PIP,
        )
    elif image_dep_file and (image_dep_file.lower().endswith('.yaml') or image_dep_file.lower().endswith('.yml')):
        dep_file_info = V1DependencyFileInfo(
            path=image_dep_file,
            package_manager=V1PackageManager.CONDA,
        )
    else:
        dep_file_info = None

    datastore_mounts = None
    if datastores:
        datastore_mounts = [V1DatastoreMounts(id=key, mount_path=value) for key, value in datastores.items()]

    request = V1CreateRunRequest(
        name=run_name,
        description=run_description,
        local_dir=localdir,
        spec=V1RunSpec(
            actions=V1RunActions(
                on_build=on_build_actions,
                on_build_start=on_build_start_actions,
                on_build_end=on_build_end_actions,
                on_experiment_start=on_experiment_start_actions,
                on_experiment_end=on_experiment_end_actions,
            ),
            cluster_id=cluster_id,
            datastores=datastore_mounts,
            desired_state=V1RunState.SUCCEEDED,
            dry_run=dry_run,
            env=per_exp_env_vars,
            image=Externalv1ImageSpec(
                dependency_file_info=dep_file_info,
                dockerfile=image_dockerfile,
                framework=image_framework,
            ),
            instance_type=instance_type,
            relative_work_dir=relative_work_dir,
            resources=V1Resources(
                cpu=per_exp_resources_cpu,
                gpu=per_exp_resources_gpu,
                storage_gb=per_exp_resources_storage_gb,
                extra=per_exp_resources_extra,
            ),
            run_controller_command=run_controller_command,
            scratch=[Externalv1ScratchSpace(mount_path=scratch_mount_path, size_gb=scratch_size_gb)],
            source_code=source_code,
            sweep_options=run_sweep_options,
            sweep_type=run_sweep_type,
            use_spot=use_spot,
            auto_resume=auto_resume,
        ),
    )
    return client.run_service_create_run(spec_cluster_id=cluster_id, body=request)


@throw_with_message
def cancel_run(client: GridRestClient, cluster_id: str, run_id: str) -> V1Run:
    run = get_run_from_id(client=client, cluster_id=cluster_id, run_id=run_id)
    run_spec = run.spec
    run_spec.desired_state = V1RunState.CANCELED
    return client.run_service_update_run(
        spec_cluster_id=cluster_id, id=run_id, body=Body8(name=run.name, spec=run_spec)
    )


@throw_with_message
def delete_run(client: GridRestClient, cluster_id: str, run_id: str) -> V1DeleteRunResponse:
    return client.run_service_delete_run(cluster_id=cluster_id, id=run_id)


@throw_with_message
def get_run_from_id(client: GridRestClient, cluster_id: str, run_id: str) -> V1Run:
    return client.run_service_get_run(cluster_id=cluster_id, id=run_id)


@throw_with_message
def get_run_from_name(client: GridRestClient, cluster_id: str, run_name: str) -> V1Run:
    runs = list_runs(client=client, cluster_id=cluster_id, query=run_name)
    for run in runs:
        if run.name == run_name:
            return run
    raise KeyError(f'Run with name {run_name} not found')


@throw_with_message
def list_runs(client: GridRestClient, cluster_id: str, user_ids=[], query: str = None) -> List[V1Run]:
    phase_not_in = [V1RunState.DELETED]
    kwargs = dict(cluster_id=cluster_id, user_ids=user_ids, phase_not_in=phase_not_in)
    if query:
        kwargs['query'] = query

    resp = client.run_service_list_runs(**kwargs)
    return resp.runs



from typing import Optional

import click

from grid.cli import rich_click
import grid.sdk.env as env
from grid.sdk.login import login as grid_login
from grid.sdk.rest.exceptions import GridException


@click.option('--username', type=str, help='Username used in Grid')
@click.option('--key', type=str, help='API Key from Grid')
@rich_click.command()
def login(username: Optional[str] = None, key: Optional[str] = None) -> None:
    """Authorize the CLI to access Grid AI resources for a particular user.

    If no username or key is provided, the CLI will prompt for them. After
    providing your username, a web browser will open to your account settings
    page where your API key can be found.
    """
    # Prompt the user for username and API Key.
    # We ask for the username first because that's
    # the same as their Github usernames.
    if not username:
        username = click.prompt('Please provide your Grid or GitHub username')
    if not key:
        settings_url = env.GRID_URL + "#/settings?tabId=apikey"
        click.launch(settings_url)
        key = click.prompt('Please provide your Grid API key')

    try:
        status = grid_login(username=username, api_key=key)
    except ConnectionRefusedError:  # only relevant until we have gQL
        raise click.ClickException('Invalid username or API key')
    except GridException as e:
        raise click.ClickException(str(e))
    if status:
        click.echo('Login successful. Welcome to Grid.')



import os
import shutil
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Optional, List

import click
from rich.console import Console
from rich.prompt import Confirm
from yaspin import yaspin

from grid.cli import rich_click
from grid.cli.observables import BaseObservable
from grid.cli.rich_click import deprecate_option
from grid.sdk import env
from grid.sdk.client import create_swagger_client
from grid.sdk.datastores import Datastore, list_datastores
from grid.sdk.rest import GridRestClient
from grid.sdk.user_messaging import info, questions
from grid.sdk.utils.datastore_uploads import (
    resume_datastore_upload,
    find_incomplete_datastore_upload,
    load_datastore_work_state,
    remove_datastore_work_state,
)

WARNING_STR = click.style('WARNING', fg='yellow')


@rich_click.group(invoke_without_command=True)
@click.pass_context
@click.option(
    '--global',
    'is_global',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help='Fetch sessions from everyone in the team when flag is passed'
)
@click.option(
    '--cluster',
    'cluster_id',
    type=str,
    required=False,
    default=env.CONTEXT,
    show_default=True,
    help='The cluster id to list datastores for.',
)
@click.option(
    '--show-incomplete',
    'show_incomplete',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help=(
        'Show any datastore uploads which were started, but killed or errored before '
        'they finished uploading all data and became "viewable" on the grid datastore '
        'user interface.'
    )
)
def datastore(ctx, cluster_id: str, is_global: bool, show_incomplete: bool) -> None:
    """Manages Datastore workflows."""
    if ctx.invoked_subcommand is None:

        table_rows, table_cols = [], []
        if show_incomplete is True:
            # TODO: Why are we using yaspin when Rich already contains a spinner console instance?
            #       This is literally an additional dependency for no reason...
            spinner = yaspin(text=f"Loading Incomplete Datastores on Local Machine...", color="yellow")
            spinner.start()

            table_cols = ["Name", "Cluster ID", "Started At", "Source Path"]
            try:
                incomplete_id = find_incomplete_datastore_upload(grid_dir=Path(env.GRID_DIR))
                if incomplete_id is not None:
                    ds = load_datastore_work_state(grid_dir=Path(env.GRID_DIR), datastore_id=incomplete_id)
                    created = f"{datetime.fromtimestamp(ds.creation_timestamp):%Y-%m-%d %H:%M}"
                    table_rows.append([ds.datastore_name, ds.cluster_id, created, ds.source])
            except Exception as e:
                spinner.fail("✘")
                raise click.ClickException(e)
        else:
            spinner = yaspin(text=f"Loading Datastores in {env.CONTEXT}...", color="yellow")
            spinner.start()

            try:
                datastores: List[Datastore] = list_datastores(cluster_id=cluster_id, is_global=is_global)
            except Exception as e:
                spinner.fail("✘")
                raise click.ClickException(str(e))

            if is_global is True:
                table_cols = [
                    "Name", "Cluster ID", "Version", "Size", "Created At", "Created By", "Team Name", "Status"
                ]
                for ds in sorted(datastores, key=lambda k: (k.name, k.version)):
                    created = f'{ds.created_at:%Y-%m-%d %H:%M}'
                    size = ds.size
                    owner = ds.user.username
                    team = ds.user.team_name
                    status = ds.status
                    table_rows.append([ds.name, ds.cluster_id, str(ds.version), size, created, owner, team, status])
            else:
                table_cols = ["Name", "Cluster ID", "Version", "Size", "Created At", "Status"]
                for ds in sorted(datastores, key=lambda k: (k.name, k.version)):
                    created = f'{ds.created_at:%Y-%m-%d %H:%M}'
                    size = ds.size
                    status = ds.status
                    table_rows.append([ds.name, ds.cluster_id, str(ds.version), size, created, status])

        table = BaseObservable.create_table(columns=table_cols)
        for row in table_rows:
            table.add_row(*row)

        spinner.ok("✔")
        console = Console()
        console.print(table)

    elif is_global:
        click.echo(f"{WARNING_STR}: --global flag doesn't have any effect when invoked with a subcommand")


@datastore.command()
@click.pass_context
def resume(ctx):
    """Resume uploading an incomplete datastore upload session."""
    _resume()


def _resume():
    """Implementation of the resume_command function not requiring the context to be passed.
    """
    Console()  # THIS IS IMPORTANT! (otherwise console text overwrites eachother)
    with Console().status("[bold yellow]Indexing incomplete datastore uploads....", spinner_style="yellow") as status:
        incomplete_id = find_incomplete_datastore_upload(grid_dir=Path(env.GRID_DIR))
        if incomplete_id is not None:
            try:
                ds_work = load_datastore_work_state(grid_dir=Path(env.GRID_DIR), datastore_id=incomplete_id)
                status.update("[bold yellow]Checking for modified files on disk...")
                modified_files = ds_work.check_for_modified_files()
                status.stop()
                if len(modified_files) > 0:
                    should_resume = Confirm.ask(
                        questions.datastore_ask_if_should_resume_after_files_modified(modified_files),
                        default=True,
                    )
                    if not should_resume:
                        click.echo("Exiting!")
                        return
                c = GridRestClient(create_swagger_client())
                click.echo(f"creating datastore from {ds_work.source}")
                resume_datastore_upload(client=c, grid_dir=Path(env.GRID_DIR), work=ds_work)
            except Exception as e:
                raise click.ClickException(e)
        else:
            raise click.ClickException(f"no incomplete datastore upload sessions exist")


@datastore.command()
@click.pass_context
def clearcache(ctx) -> None:
    """Clears datastore cache which is saved on the local machine when uploading a datastore to grid.

    This removes all the cached files from the local machine, meaning that resuming an incomplete
    upload is not possible after running this command.
    """
    grid_dstore_path = Path(env.GRID_DIR).joinpath("datastores")
    grid_dstore_path.mkdir(parents=True, exist_ok=True)
    for f in grid_dstore_path.iterdir():
        if f.is_file():
            os.remove(str(f.absolute()))
        if f.is_dir():
            shutil.rmtree(str(f.absolute()))
    click.echo("Datastore cache cleared")


@datastore.command(cls=rich_click.deprecate_grid_options())
@click.option('--name', type=str, required=True, help='Name of the datastore')
@click.option('--version', type=int, required=True, help='Version of the datastore')
@click.option(
    '--cluster',
    type=str,
    required=False,
    default=env.CONTEXT,
    show_default=True,
    help='cluster id to delete the datastore from. (Bring Your Own Cloud Customers Only).'
)
@click.pass_context
def delete(ctx, name: str, version: int, cluster: str) -> None:
    """Deletes a datastore with the given name and version tag.

    For bring-your-own-cloud customers, the cluster id of the associated
    resource is required as well.
    """
    try:
        dstore = Datastore(name=name, version=version, cluster_id=cluster)
        dstore.delete()
    except Exception as e:
        raise click.ClickException(str(e))
    click.echo("Done!")


@datastore.command()
@rich_click.argument(
    'source',
    type=str,
    default=None,
    required=False,
    help=(
        "Source to create datastore from. This could either be a local "
        "directory (e.g: /opt/local_folder) a remote http URL pointing "
        "to a TAR or ZIP file (e.g. http://some_domain/data.tar.gz), or "
        "an s3 bucket to copy data from (e.g. s3://ryft-public-sample-data/esRedditJson/)"
    )
)
@click.option(
    '--source',
    'source_',
    type=str,
    show_default=False,
    default=None,
    required=False,
    hidden=True,
    callback=partial(deprecate_option, "the argument form (grid datastore create SOURCE) of this command"),
)
@click.option(
    "--no-copy",
    "_no_copy_data_source",
    is_flag=True,
    default=False,
    help=(
        "(beta) Use this flag when you intend to incrementally add data to the source bucket. "
        "Using this flag can also speed up datastore creation when working with large buckets. "
        "When using this flag, you cannot remove files from your bucket. If you'd like to add "
        "files, please create a new version of the datastore after you've added files to your "
        "bucket. Please note that Grid does not currently support private S3 buckets."
    ),
)
@click.option('--name', type=str, required=False, help='Name of the datastore')
@click.option(
    '--cluster',
    type=str,
    default=env.CONTEXT,
    show_default=True,
    required=False,
    help='cluster id to create the datastore on. (Bring Your Own Cloud Customers Only).'
)
@click.option(
    "--hpd",
    "_fsx_enabled",
    is_flag=True,
    default=False,
    help=(
        "(beta) Use this flag to provision a HPD datastore backed by AWS FSx for Lustre. This "
        "type of datastore is automatically updated whenever new files are added or deleted from the source "
        "S3 bucket. Please take this into account when creating workflows around such Datastores. \n\n"
        "This feature is only available to BYOC customers. "
    ),
)
@click.option(
    "--hpd-throughput",
    "_fsx_throughput_alias",
    type=str,
    default="low",
    help=(
        "(beta) Throughput setting for HPDs. Select one of [low, medium, high]. \n\n"
        "low (default): 125mb/s of throughput per tib of Datastore capacity. Recommended for Datastores "
        "that will be used by one or two experiments or sessions simultaneously. \n\n"
        "medium: 500mb/s per tib of Datastore capacity. Recommended for Datastores that will be used by "
        "multiple experiments or sessions simultaneously. \n\n"
        "high: highest possible throughput of 1000mb/s per tib of Datastore capacity. Recommended when "
        "maximum performance is necessary to run a very high number of experiments. \n\n"
        "Please note that for a single session or experiment, selecting the medium or "
        "high settings will yield diminishing returns. In these cases, we recommended "
        "selecting the low (default) throughput option. \n\n"
        "Because HPDs offer elevated performance, please note that they can incur higher monthly "
        "costs than regular Grid Datastores, especially for the medium or high throughput options. \n\n"
        "This feature is only available to BYOC customers. "
    ),
)
@click.option(
    "--hpd-capacity",
    "_fsx_capacity_gib",
    type=int,
    default=1200,
    help=(
        "(beta) Capacity setting for HPDs in GiB. Must be 1200, 2400 or "
        "a multiple of 2400, up to 64800. \n\n"
        "This feature is only available to BYOC customers. "
    ),
)
@click.option(
    "--hpd-preload",
    "_fsx_preloading",
    type=bool,
    is_flag=True,
    default=False,
    help=(
        "(beta) Use this flag when provisioning a high-performance datastore "
        "when maximum performance is needed even on the first data access. (Due to "
        "technical reasons HPDs without this flag achieve maximum performance after "
        "the first data access). Avoid using this flag when the datastore needs to be "
        "ready for use as soon as possible or when you're only working on a partition "
        "of the data in the S3 bucket. \n\n"
        "This feature is only available to BYOC customers. "
    ),
)
@click.pass_context
def create(
    ctx,
    source: str,
    source_: str,
    cluster: str,
    _no_copy_data_source: bool,
    _fsx_enabled: bool,
    _fsx_throughput_alias: str,
    _fsx_capacity_gib: int,
    _fsx_preloading: bool,
    name: Optional[str] = None
) -> None:
    """Creates a datastore from SOURCE.

    If you want to check the status of your datastore creation, please use the
    `grid datastore` command.

    Here are some examples of this command in use:

    To create a datastore from a directory (or file) on the local machine
    (optionally specifying a name):

        grid datastore create ./my_cool_file.txt

        grid datastore create ./some-directory/

        grid datastore create ./some-directory/ --name my-awesome-datastore

    To create a datastore from an S3 bucket (private buckets are not currently
    supported):

        grid datastore create s3://ryft-public-sample-data/esRedditJson/

    If you'd like to create a datastore from an S3 bucket that will be
    incrementally updated, or for very large datasets:

        grid datastore create s3://ryft-public-sample-data/ --no-copy

    To create a datastore from a zip or tar.gz file hosted at a URL:

        grid datastore create https://cs.nyu.edu/~roweis/data/EachMovieData.tar.gz
    """

    if source_ is not None:
        source = source_
    try:
        Console()  # THIS IS IMPORTANT! (otherwise console text overwrites eachother)
        incomplete_id = find_incomplete_datastore_upload(Path(env.GRID_DIR))
        if incomplete_id is not None:
            should_resume = Confirm.ask(
                questions.datastore_ask_if_should_resume_upload(source),
                default=True,
            )
            if should_resume is True:
                _resume()
            else:
                remove_datastore_work_state(grid_dir=Path(env.GRID_DIR), datastore_id=incomplete_id)
                dstore = Datastore(
                    name=name,
                    source=source,
                    cluster_id=cluster,
                    s3_no_copy=_no_copy_data_source,
                    fsx_enabled=_fsx_enabled,
                    fsx_throughput_alias=_fsx_throughput_alias,
                    fsx_capacity_gib=_fsx_capacity_gib,
                    fsx_preloading=_fsx_preloading,
                )
                dstore.upload()
        else:
            with Console().status("[bold yellow]Indexing datastore uploads....", spinner_style="yellow") as status:
                dstore = Datastore(
                    name=name,
                    source=source,
                    cluster_id=cluster,
                    s3_no_copy=_no_copy_data_source,
                    fsx_enabled=_fsx_enabled,
                    fsx_throughput_alias=_fsx_throughput_alias,
                    fsx_capacity_gib=_fsx_capacity_gib,
                    fsx_preloading=_fsx_preloading,
                )
                status.stop()
                dstore.upload()
    except Exception as e:
        raise click.ClickException(e)

    # Data store created confirmation message
    click.echo("Creating Datastore...")
    Console().print(info.datastore_uploaded_info_table(dstore))
    click.echo("Please use `grid datastore` to check the status.")



from pathlib import Path
import subprocess
from typing import List, Union

import grid.sdk.env as env


def execute_git_command(args: List[str], cwd=None) -> str:
    """
    Executes a git command. This is expected to return a
    single string back.

    Returns
    -------
    output: str
        String combining stdout and stderr.
    """
    process = subprocess.run(['git'] + args,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=True,
                             cwd=None,
                             check=False)

    output = process.stdout.strip() + process.stderr.strip()
    return output


def check_if_remote_head_is_different() -> Union[bool, None]:
    """
    Checks if remote git repository is different than
    the version available locally. This only compares the
    local SHA to the HEAD commit of a given branch. This
    check won't be used if user isn't in a HEAD locally.

    Original solution:

        * https://stackoverflow.com/questions/\
            3258243/check-if-pull-needed-in-git
    """
    # Check SHA values.
    local_sha = execute_git_command(['rev-parse', '@'])
    remote_sha = execute_git_command(['rev-parse', r"@{u}"])
    base_sha = execute_git_command(['merge-base', '@', r"@{u}"])

    # Whenever a SHA is not avaialble, just return.
    if any('fatal' in f for f in (local_sha, remote_sha, base_sha)):
        return None

    is_different = True
    if local_sha in (remote_sha, base_sha):
        is_different = False

    return is_different


def _check_github_repository() -> None:
    """Checks if the active directory is a GitHub repository."""
    github_repository = execute_git_command(["config", "--get", "remote.origin.url"])

    if not github_repository or 'github.com' not in github_repository:
        # TODO - change error message to advertise `--localdir`
        raise RuntimeError(
            '`grid train` or `grid interactive` can only be run in a git repository '
            'hosted on github.com. See docs for details: https://docs.grid.ai'
        )


def check_if_uncommited_files() -> bool:
    """
    Checks if user has uncommited files in local repository.
    If there are uncommited files, then show a prompt
    indicating that uncommited files exist locally.

    Original solution:

        * https://stackoverflow.com/questions/3878624/how-do-i-programmatically-determine-if-there-are-uncommitted-changes
    """
    files = execute_git_command(['update-index', '--refresh'])
    return bool(files)


def add_git_root_path(file: Union[str, Path]) -> str:
    #  Finds the relative path of the file to train.
    abs_path = Path(file).absolute()
    repository_path = execute_git_command(['rev-parse', '--show-toplevel'], cwd=abs_path)
    return str(abs_path.relative_to(repository_path))


def check_github_repository() -> None:
    """Checks if the active directory is a GitHub repository."""
    github_repository = execute_git_command(["config", "--get", "remote.origin.url"])
    env.logger.debug(github_repository)

    if not github_repository or 'github.com' not in github_repository:
        # TODO - change error message to advertise `--localdir`
        raise ValueError(
            '`grid train` or `grid interactive` can only be run in a git repository '
            f'hosted on github.com. found remote origin url={github_repository}. '
            f'See docs for details: https://docs.grid.ai'
        )



from dataclasses import dataclass
import json
import os
from pathlib import Path
from typing import Union

from grid.sdk import env


@dataclass(unsafe_hash=True)
class Credentials:
    user_id: str
    api_key: str

    @classmethod
    def from_locale(cls) -> "Credentials":
        """Instantiates the credentials using implicit locale.

        First use environment variables, otherwise look for credentials stored in file.

        Returns
        -------
        Credentials
            instantiated credentials object.
        """
        # if user has environment variables, use that
        user_id = os.getenv('GRID_USER_ID')
        api_key = os.getenv('GRID_API_KEY')
        grid_url = os.getenv('GRID_URL')
        if grid_url:
            env.GRID_URL = grid_url
        if user_id and api_key:
            return cls(user_id=user_id, api_key=api_key)

        # otherwise overwrite look for credentials stored locally as a file
        if os.getenv("CI"):
            p = Path.home() / ".grid" / "credentials.json"
        else:
            p = Path(os.getenv('GRID_CREDENTIAL_PATH', Path.home() / ".grid" / "credentials.json"))

        if not p.exists():
            raise PermissionError('No credentials available. Did you login?')
        with p.open() as f:
            credentials = json.load(f)

        return cls(
            user_id=credentials['UserID'],
            api_key=credentials['APIKey'],
        )

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> "Credentials":
        """Instantiates the credentials using specified file path.

        Parameters
        ----------
        path
            file path to the config yaml or json on disk.

        Returns
        -------
        Credentials
            instantiated credentials object.
        """
        p = Path(path).absolute()
        if not p.exists():
            raise PermissionError('No credentials available. Did you login?')
        with p.open() as f:
            credentials = json.load(f)

        return cls(
            user_id=credentials['UserID'],
            api_key=credentials['APIKey'],
        )
